// <copyright file="registrator.cpp" company="Aalborg University">
// Copyright (c) 2014 All Right Reserved, http://vap.aau.dk/
//
// This source is subject to the MIT Licence
// Copyright (c) <2014> <Aalborg University>
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.
//
// </copyright>
// <author>Chris Bahnsen</author>
// <email>cb@create.aau.dk</email>
// <email2>chris.bahnsen@hotmail.com</email2>
// <date>2014-11-10</date>
// <summary>Contains the base functions for registering trimodal imagery by use of a calibVars.yml function</summary>

#include "registrator.h"

using namespace std;
using namespace cv;

Registrator::Registrator()
{
    // Initialize variables where no sensible default exists
    stereoCalibParam.defaultDepth = 0;
    stereoCalibParam.depthCoeffA = 0;
    stereoCalibParam.depthCoeffB = 0;
    stereoCalibParam.HEIGHT = 0;
    stereoCalibParam.nbrHom = 0;
    stereoCalibParam.WIDTH = 0;

    settings.isInitialized = false;

}

float Registrator::backProjectPoint(float point, float focalLength, float principalPoint, float zCoord)
{
    float projectedPoint;

    projectedPoint = (point * zCoord - principalPoint * zCoord) / (focalLength);
    return projectedPoint;
}

float Registrator::forwardProjectPoint(float point, float focalLength, float principalPoint, float zCoord)
{
    float projectedPoint;

    projectedPoint = (point * focalLength) / zCoord + principalPoint;
    return projectedPoint;
}

void Registrator::readCoordinatesFromFile2D(string file, vector<vector<Point2f> >& imagePoints)
{
    string lineX, lineY, lineZ;
    char buffX[20], buffY[20];
    float f;
    vector<Point2f> tempVector;
    Point2f tempPoint;
    ifstream fileH(file);
    int count = 1;

    if (fileH.is_open())
    {
        while (fileH.good())
        {
            //cout << count++;

            // Clear temporary variables
            lineX.clear(); lineY.clear(); lineZ.clear(), tempVector.clear();

            getline(fileH, lineX);
            getline(fileH, lineY);
            getline(fileH, lineZ);
            stringstream subStrX, subStrY;
            subStrX << lineX;
            subStrY << lineY;

            while ((subStrX.good() && subStrY.good()))
            {
                // Get all the entries in one line
                subStrX.getline(buffX, 12, ',');
                subStrY.getline(buffY, 12, ',');

                if (EOF == sscanf(buffX, "%f", &f))
                {
                    // Error
                }
                else
                {
                    // Place float value in point array
                    tempPoint.x = f;
                    sscanf(buffY, "%f", &f);
                    tempPoint.y = f;

                    // Place point array in vector
                    tempVector.push_back(tempPoint);
                }
            }
            // Place vector in vector
            imagePoints.push_back(tempVector);
        }
        fileH.close();
    }
    else cout << "Unable to open file";

    return;
}

void Registrator::readCoordinatesFromFile3D(string file, vector<vector<Point3f> >& objectPoints)
{
    string lineX, lineY, lineZ, lineDel;
    char buffX[20], buffY[20], buffZ[20];
    float f;
    vector<Point3f> tempVector;
    Point3f tempPoint;
    ifstream fileH(file);

    if (fileH.is_open())
    {
        while (fileH.good())
        {
            // Clear temporary variables
            lineX.clear(); lineY.clear(); lineZ.clear(), tempVector.clear();

            getline(fileH, lineX);
            getline(fileH, lineY);
            getline(fileH, lineZ);
            getline(fileH, lineDel);
            stringstream subStrX, subStrY, subStrZ;
            subStrX << lineX;
            subStrY << lineY;
            subStrZ << lineZ;

            while ((subStrX.good() && subStrY.good() && subStrZ.good()))
            {
                // Get all the entries in one line
                subStrX.getline(buffX, 12, ',');
                subStrY.getline(buffY, 12, ',');
                subStrZ.getline(buffZ, 12, ',');

                if (EOF == sscanf(buffX, "%f", &f))
                {
                    // Error
                }
                else
                {
                    // Place float value in point array
                    tempPoint.x = f;
                    sscanf(buffY, "%f", &f);
                    tempPoint.y = f;
                    sscanf(buffZ, "%f", &f);
                    tempPoint.z = f;

                    // Place point array in vector
                    tempVector.push_back(tempPoint);
                }
            }
            // Place vector in vector
            objectPoints.push_back(tempVector);
        }
        fileH.close();
    }
    else cout << "Unable to open file";

    return;
}

float Registrator::lookUpDepth(Mat depthImg, Point2f dCoord, bool SCALE_TO_THEORETICAL)
{
    float depthInMm = 0;
    // Look up the 'vanilla' depth in the depth image


    if (!depthImg.empty() && (dCoord.y > 0 && dCoord.y < depthImg.rows
        && dCoord.x > 0 && dCoord.x < depthImg.cols)) {
        depthInMm = depthImg.at<unsigned short>(int(dCoord.y - 1), int(dCoord.x - 1))*0.1;
    }
    else {
        depthInMm = stereoCalibParam.defaultDepth;
    }

    if (SCALE_TO_THEORETICAL)
    {
        // Transfer the depth to the "theoretical" depth range as provided by the camera calibration
        depthInMm = stereoCalibParam.depthCoeffA*depthInMm + stereoCalibParam.depthCoeffB;
    }

    return depthInMm;
}

void Registrator::computeCorrespondingRgbPointFromDepth(vector<Point2f> vecDCoord, vector<Point2f> & vecRgbCoord)
{
    vector<Point2f> vecDistRgbCoord, vecUndistRgbCoord, vecRecRgbCoord;
    Point2f tmpPoint;
    Point2f prevPoint;

    // Find the distorted rgb point
    for (size_t i = 0; i < vecDCoord.size(); i++)
    {
        if ((vecDCoord[i].y > 0) && (vecDCoord[i].x > 0) && (vecDCoord[i].y <= stereoCalibParam.HEIGHT) && (vecDCoord[i].x <= stereoCalibParam.WIDTH)) {
            tmpPoint.x = stereoCalibParam.dToRgbCalX.at<short>(int(vecDCoord[i].y - 1), int(vecDCoord[i].x - 1)); // Look up the corresponding x-point in the depth image
            tmpPoint.y = stereoCalibParam.dToRgbCalY.at<short>(int(vecDCoord[i].y - 1), int(vecDCoord[i].x - 1)); // Look up the corresponding y-point in the depth image

            if ((tmpPoint.x == 0) || (tmpPoint.y == 480) || (tmpPoint.y == 0))
            {
                // If the point is inside the frame of the depth camera, look up neighbouring points and perform bilinear interpolation
                Point2f interpolPoint(0, 0), tmpLookupPoint;;
                int lookupCount = 0, xOffset, yOffset;

                for (int j = 0; j < 4; ++j) {

                    switch (j) {
                    case 0:
                        xOffset = -1;
                        yOffset = -1;
                        break;
                    case 1:
                        xOffset = +1;
                        yOffset = -1;
                        break;
                    case 2:
                        xOffset = +1;
                        yOffset = +1;
                        break;
                    case 3:
                        xOffset = -1;
                        yOffset = +1;
                        break;
                    }

                    if (((vecDCoord[i].y + yOffset) > 0) && ((vecDCoord[i].x + xOffset) > 0) && ((vecDCoord[i].y + yOffset) <= stereoCalibParam.HEIGHT)
                        && ((vecDCoord[i].x + xOffset) <= stereoCalibParam.WIDTH)) {
                        tmpLookupPoint.x = stereoCalibParam.dToRgbCalX.at<short>(int(vecDCoord[i].y - 1 + yOffset), int(vecDCoord[i].x - 1 + xOffset)); // Look up the corresponding x-point in the depth image
                        tmpLookupPoint.y = stereoCalibParam.dToRgbCalY.at<short>(int(vecDCoord[i].y - 1 + yOffset), int(vecDCoord[i].x - 1 + xOffset)); // Look up the corresponding y-point in the depth image

                        if ((tmpLookupPoint.x != 0) && (tmpLookupPoint.y != 480) && (tmpLookupPoint.y != 0)) {
                            interpolPoint.x = interpolPoint.x + tmpLookupPoint.x;
                            interpolPoint.y = interpolPoint.y + tmpLookupPoint.y;
                            lookupCount++;
                        }
                    }
                }

                if (lookupCount > 0) {
                    tmpPoint.x = interpolPoint.x / lookupCount;
                    tmpPoint.y = interpolPoint.y / lookupCount;
                }
                else if (this->settings.USE_PREV_DEPTH_POINT)
                {
                    tmpPoint = prevPoint;
                }
                else {
                    tmpPoint = Point2f(0, 0);
                }
            }
            vecDistRgbCoord.push_back(tmpPoint);
        }
        else {
            tmpPoint.x = 1; tmpPoint.y = 1;
            vecDistRgbCoord.push_back(tmpPoint);
        }
    }

    if (settings.UNDISTORT_IMAGES) {
        undistortPoints(vecDistRgbCoord, vecUndistRgbCoord, stereoCalibParam.rgbCamMat, stereoCalibParam.rgbDistCoeff,
            Mat::eye(3, 3, CV_32F), stereoCalibParam.rgbCamMat);
    }

    if (settings.UNDISTORT_IMAGES) {
        // If images are undistorted, but not rectified, use undistorted coordinates
        vecRgbCoord = vecUndistRgbCoord;
    }
    else {
        // Otherwise, use the distorted coordinates
        vecRgbCoord = vecDistRgbCoord;
    }

}

void Registrator::computeCorrespondingDepthPointFromRgb(vector<Point2f> vecRgbCoord, vector<Point2f> & vecDCoord)
{
    vector<Point2f> vecDistRgbCoord, vecUndistRgbCoord;
    Point2f tmpPoint;
    Point2f prevPoint(0, 0);

    // Now, we may re-distort the points so they correspond to the original image
    if (settings.UNDISTORT_IMAGES) {
        // If we need to undistort the point, do so:
        MyDistortPoints(vecRgbCoord, vecDistRgbCoord, stereoCalibParam.rgbCamMat, stereoCalibParam.rgbDistCoeff);
    }
    else {
        // Otherwise, the point is already distorted
        vecDistRgbCoord = vecRgbCoord;
    }


    if ((stereoCalibParam.rgbToDCalX.cols == stereoCalibParam.WIDTH) && (stereoCalibParam.rgbToDCalX.rows == stereoCalibParam.HEIGHT))
    {
        for (size_t i = 0; i < vecDistRgbCoord.size(); i++)
        {
            if ((vecDistRgbCoord[i].y > 0) && (vecDistRgbCoord[i].x > 0) && (vecDistRgbCoord[i].y <= stereoCalibParam.HEIGHT)
                && (vecDistRgbCoord[i].x <= stereoCalibParam.WIDTH)) {
                tmpPoint.x = stereoCalibParam.rgbToDCalX.at<short>(int(vecDistRgbCoord[i].y - 1), int(vecDistRgbCoord[i].x - 1)); // Look up the corresponding x-point in the depth image
                tmpPoint.y = stereoCalibParam.rgbToDCalY.at<short>(int(vecDistRgbCoord[i].y - 1), int(vecDistRgbCoord[i].x - 1)); // Look up the corresponding y-point in the depth image

                if ((tmpPoint.x == 0) || (tmpPoint.y == 480) || (tmpPoint.y == 0)) {
                    // If the point is inside the frame of the RGB camera, look up neighbouring points and perform bilinear interpolation
                    Point2f interpolPoint(0, 0), tmpLookupPoint;;
                    int lookupCount = 0, xOffset, yOffset;

                    for (int j = 0; j < 4; ++j) {

                        switch (j) {
                        case 0:
                            xOffset = -1;
                            yOffset = -1;
                            break;
                        case 1:
                            xOffset = +1;
                            yOffset = -1;
                            break;
                        case 2:
                            xOffset = +1;
                            yOffset = +1;
                            break;
                        case 3:
                            xOffset = -1;
                            yOffset = +1;
                            break;
                        }

                        if (((vecDistRgbCoord[i].y + yOffset) > 0) && ((vecDistRgbCoord[i].x + xOffset) > 0) && ((vecDistRgbCoord[i].y + yOffset) <= stereoCalibParam.HEIGHT)
                            && ((vecDistRgbCoord[i].x + xOffset) <= stereoCalibParam.WIDTH)) {
                            tmpLookupPoint.x = stereoCalibParam.rgbToDCalX.at<short>(int(vecDistRgbCoord[i].y - 1 + yOffset), int(vecDistRgbCoord[i].x - 1 + xOffset)); // Look up the corresponding x-point in the depth image
                            tmpLookupPoint.y = stereoCalibParam.rgbToDCalY.at<short>(int(vecDistRgbCoord[i].y - 1 + yOffset), int(vecDistRgbCoord[i].x - 1 + xOffset)); // Look up the corresponding y-point in the depth image

                            if ((tmpLookupPoint.x != 0) && (tmpLookupPoint.y != 480) && (tmpLookupPoint.y != 0)) {
                                interpolPoint.x = interpolPoint.x + tmpLookupPoint.x;
                                interpolPoint.y = interpolPoint.y + tmpLookupPoint.y;
                                lookupCount++;
                            }
                        }
                    }

                    if (lookupCount > 0) {
                        tmpPoint.x = interpolPoint.x / lookupCount;
                        tmpPoint.y = interpolPoint.y / lookupCount;
                    }
                    else if (this->settings.USE_PREV_DEPTH_POINT)
                    {
                        tmpPoint = prevPoint;
                    }
                    else {
                        tmpPoint = Point2f(0, 0);
                    }
                }
                vecDCoord.push_back(tmpPoint);
            }
            else {
                if (this->settings.USE_PREV_DEPTH_POINT)
                {
                    // If the RGB coordinates are not within the camera frame, return the previous value
                    vecDCoord.push_back(prevPoint);
                }
                else {
                    vecDCoord.push_back(Point2f(0, 0));
                }
            }

            prevPoint = tmpPoint;
        }
    }

}


void Registrator::computeCorrespondingThermalPointFromRgb(vector<Point2f> vecRgbCoord, vector<Point2f>& vecTCoord, vector<Point2f> vecDCoord)
{
    vector<int> vecDepthInMm, bestHom;
    vector<double> minDist;
    vector<vector<double> > octantDistances;
    vector<vector<int> > octantIndices;
    vector<Point3f> worldCoordPointVector;
    computeCorrespondingThermalPointFromRgb(vecRgbCoord, vecTCoord, vecDCoord, vecDepthInMm, minDist,
        bestHom, octantIndices, octantDistances, worldCoordPointVector);

}

void Registrator::computeCorrespondingThermalPointFromRgb(vector<Point2f> vecRgbCoord, vector<Point2f>& vecTCoord, vector<Point2f> vecDCoord,
    vector<int> &bestHom)
{
    vector<int> vecDepthInMm;
    vector<double> minDist;
    vecDepthInMm.resize(0);
    vector<vector<double> > octantDistances;
    vector<vector<int> > octantIndices;
    vector<Point3f> worldCoordPointVector;
    computeCorrespondingThermalPointFromRgb(vecRgbCoord, vecTCoord, vecDCoord, vecDepthInMm, minDist,
        bestHom, octantIndices, octantDistances, worldCoordPointVector);
}

void Registrator::computeCorrespondingThermalPointFromRgb(vector<Point2f> vecRgbCoord, vector<Point2f>& vecTCoord, vector<Point2f> vecDCoord,
    vector<int> vecDepthInMm, vector<double>& minDist, vector<int> &bestHom, vector<vector<int> > &octantIndices,
    vector<vector<double> > &octantDistances, vector<Point3f> &worldCoordPointVector)
{
    vector<Point2f> vecUndistRgbCoord, vecRecRgbCoord, vecDistRgbCoord, vecRecTCoord, vecUndistTCoord;
    Point2f tmpPoint;
    minDist.clear(); bestHom.clear();

    if (!settings.UNDISTORT_IMAGES && (stereoCalibParam.rgbCamMat.size() == cv::Size(3, 3))) // If the images are not undistorted, produce undistorted coordinates
    {
        undistortPoints(vecRgbCoord, vecUndistRgbCoord, stereoCalibParam.rgbCamMat, stereoCalibParam.rgbDistCoeff,
            Mat::eye(3, 3, CV_32F), stereoCalibParam.rgbCamMat);
        vecDistRgbCoord = vecRgbCoord;
    }
    else {
        vecUndistRgbCoord = vecRgbCoord;
    }

    if (settings.registrationMethod == TrimodalMultipleHom)
    {
        computeHomographyMapping(vecUndistRgbCoord, vecUndistTCoord, vecDCoord, vecDepthInMm, minDist, bestHom,
            octantIndices, octantDistances, worldCoordPointVector);
    }
    else if (settings.registrationMethod == BimodalMultipleHom)
    {
        computeHomographyMapping(vecUndistRgbCoord, vecUndistTCoord);
    }

    // If needed, distort the coordinates
    if (!settings.UNDISTORT_IMAGES && (stereoCalibParam.rgbCamMat.size() == cv::Size(3, 3))) // If the images are distorted, produce distorted coordinates
    {
        MyDistortPoints(vecUndistTCoord, vecTCoord, stereoCalibParam.tCamMat, stereoCalibParam.tDistCoeff);
    }
    else {
        vecTCoord = vecUndistTCoord;
    }

}

void Registrator::computeCorrespondingRgbPointFromThermal(vector<Point2f> vecTCoord, vector<Point2f>& vecRgbCoord)
{
    vector<double> minDist;
    vector<int> bestHom;
    vector<vector<int> > octantIndices;
    vector<vector<double> > octantDistances;
    vector<Point3f> worldCoordPointVector;

    computeCorrespondingRgbPointFromThermal(vecTCoord, vecRgbCoord, minDist, bestHom, octantIndices, octantDistances,
        worldCoordPointVector);


}

void Registrator::computeCorrespondingRgbPointFromThermal(vector<Point2f> vecTCoord, vector<Point2f>& vecRgbCoord, vector<double>& minDist,
    vector<int> &bestHom, vector<vector<int> > &octantIndices, vector<vector<double>> &octantDistances)
{
    vector<Point3f> worldCoordPointVector;
    computeCorrespondingRgbPointFromThermal(vecTCoord, vecRgbCoord, minDist, bestHom, octantIndices, octantDistances,
        worldCoordPointVector);
}

void Registrator::computeCorrespondingRgbPointFromThermal(vector<Point2f> vecTCoord, vector<Point2f>& vecRgbCoord, vector<double>& minDist,
    vector<int> &bestHom, vector<vector<int> > &octantIndices, vector<vector<double> > &octantDistances,
    vector<Point3f> &worldCoordPointVector)
{
    vector<Point2f> vecUndistTCoord, vecRecTCoord, vecDistTCoord, vecRecRgbCoord, vecUndistRgbCoord, vecDCoord;
    Point2f tmpPoint;
    minDist.clear(); bestHom.clear();

    if (!settings.UNDISTORT_IMAGES) // If the images are not undistorted, produce undistorted coordinates
    {
        undistortPoints(vecTCoord, vecUndistTCoord, stereoCalibParam.tCamMat, stereoCalibParam.tDistCoeff,
            Mat::eye(3, 3, CV_32F), stereoCalibParam.tCamMat);
        vecDistTCoord = vecTCoord;
    }
    else {
        vecUndistTCoord = vecTCoord;
    }


    if (settings.registrationMethod == TrimodalMultipleHom)
    {
        vector<int> vecDepthInMm;
        computeHomographyMapping(vecUndistRgbCoord, vecUndistTCoord, vecDCoord, vecDepthInMm, minDist, bestHom,
            octantIndices, octantDistances, worldCoordPointVector);
    }
    else if (settings.registrationMethod == BimodalMultipleHom)
    {
        computeHomographyMapping(vecUndistRgbCoord, vecUndistTCoord);
    }

    // If needed, distort the coordinates
    if (!settings.UNDISTORT_IMAGES) // If the images are distorted, produce distorted coordinates
    {
        MyDistortPoints(vecUndistRgbCoord, vecRgbCoord, stereoCalibParam.rgbCamMat, stereoCalibParam.rgbDistCoeff);
    }
    else {
        vecRgbCoord = vecUndistRgbCoord;
    }

}

void Registrator::computeHomographies(vector<Point2f> cam1Points, vector<Point2f> cam2Points, int nbrHom, vector<Mat>& homCam1Cam2,
    vector<Mat>& homCam2Cam1, vector<Point2f>& homCentersCam1, vector<Point2f>& homCentersCam2)
{
    // Normalize data before K-means
    Mat meanMat, stdDevMatX, stdDevMatY;
    vector<double> xVec, yVec;
    //Mat normPoints(cam1Points.size(), 1, CV_32FC2);
    vector<Point2f> normPoints;


    for (int i = 0; i < cam1Points.size(); i++)
    {
        // OpenCV doesn't handle the computation of the standard deviation on Matrices or vectors
        xVec.push_back(cam1Points[i].x);
        yVec.push_back(cam1Points[i].y);
    }

    meanStdDev(xVec, meanMat, stdDevMatX);
    meanStdDev(yVec, meanMat, stdDevMatY);

    Point2f tmpPoint;

    for (int i = 0; i < cam1Points.size(); i++)
    {
        tmpPoint.x = xVec[i] / stdDevMatX.at<double>(0);
        tmpPoint.y = yVec[i] / stdDevMatY.at<double>(0);
        normPoints.push_back(tmpPoint);
    }

    // Find clusters
    Mat centers(cam1Points.size(), 1, CV_32FC2), labels;

    kmeans(normPoints, nbrHom, labels,
        TermCriteria(CV_TERMCRIT_EPS, 10, 0.5),
        100, KMEANS_RANDOM_CENTERS, centers);


    vector< vector<Point2f> > clusteredCam1Points, clusteredCam2Points;
    Point3f tmp3dPoint;
    int currentCluster;

    clusteredCam1Points.resize(nbrHom);
    clusteredCam2Points.resize(nbrHom);

    // Group points into those clusters assigned by k-means
    for (int i = 0; i < cam1Points.size(); i++)
    {
        currentCluster = labels.at<int>(i);
        clusteredCam1Points[currentCluster].push_back(cam1Points[i]);
        clusteredCam2Points[currentCluster].push_back(cam2Points[i]);
    }

    // Compute planar homographies from the assigned clusters
    vector<Point2f> tmpTPoints, tmpRgbPoints;
    vector<Point3f> maskedPointCloudRgb;

    Point2f tmpImageCoordRgb, tmpImageCoordT;
    Mat tmpHom, invHom;

    homCentersCam1.clear();
    homCentersCam2.clear();

    for (int i = 0; i < nbrHom; i++)
    {
        // Compute the homographies and push them into an array
        vector<uchar> mask; Point2f meanCam1, meanCam2; int count;

        homCam1Cam2.push_back(findHomography(clusteredCam1Points[i], clusteredCam2Points[i], CV_RANSAC, 3, mask));
        invert(homCam1Cam2[i], invHom, DECOMP_SVD);
        homCam2Cam1.push_back(invHom);

        count = 0; meanCam1.x = 0; meanCam1.y = 0; meanCam2.x = 0, meanCam2.y = 0;

        // Go through the masks and find the centre of the homography plane
        for (int k = 0; k < mask.size(); k++)
        {
            if (mask[k] != 0)
            {
                count++;
                meanCam1.x = meanCam1.x + (clusteredCam1Points[i][k].x - meanCam1.x) * 1 / (count); // Iterative mean
                meanCam1.y = meanCam1.y + (clusteredCam1Points[i][k].y - meanCam1.y) * 1 / (count);

                meanCam2.x = meanCam2.x + (clusteredCam2Points[i][k].x - meanCam2.x) * 1 / (count); // Iterative mean
                meanCam2.y = meanCam2.y + (clusteredCam2Points[i][k].y - meanCam2.y) * 1 / (count);
            }
        }

        homCentersCam1.push_back(meanCam1);
        homCentersCam2.push_back(meanCam2);
    }
}

void Registrator::computeHomographyMapping(vector<Point2f>& cam1Points, vector<Point2f>& cam2Points)
{
    // Use bilinear interpolator

    double sqDist, minDistTmp, maxDistTmp;
    //Scalar sumOfWeights;

    Point tmpIdx;
    int MAP_TYPE = 0;
    int nbrCoordinates;
    int nbrClusters = stereoCalibParam.nbrHom;
    vector<Point2f> sourcePoints;

    // Are we mapping RGB or thermal points?
    if (cam1Points.size() > 0)
    {
        MAP_TYPE = 1;
        nbrCoordinates = int(cam1Points.size());
        cam2Points.clear();
        sourcePoints = cam1Points;
    }
    else if (cam2Points.size() > 0)
    {
        MAP_TYPE = 2;
        nbrCoordinates = int(cam2Points.size());
        cam1Points.clear();
        sourcePoints = cam2Points;
    }
    else
    {
        MAP_TYPE = 0;
        cout << "Error in computeHomographyMapping - no coordinates to map!";
        return;
    }

    for (int i = 0; i < nbrCoordinates; i++)
    {
        // Find the euclidean distances from the point to the homographies

        int bestHomTmp = 0; // Reset parameters
        double bestDist = 1e6;

        vector<double> homDist, homWeights;
        int errorCounter = 0;

        for (int j = 0; j < nbrClusters; j++)
        {
            double mahalDist;
            int windowRadius;
            Mat depthCenter(1, 3, CV_32F, Scalar(0));
            // Compute Mahalanobis distance, squared
            if (MAP_TYPE == 1)
            {
                homDist.push_back(sqrt(pow(sourcePoints[i].x - stereoCalibParam.homCentersCam1[j].x, 2) +
                    pow(sourcePoints[i].y - stereoCalibParam.homCentersCam1[j].y, 2)));  // Euclidean					
            }
            else {
                homDist.push_back(sqrt(pow(sourcePoints[i].x - stereoCalibParam.homCentersCam2[j].x, 2) +
                    pow(sourcePoints[i].y - stereoCalibParam.homCentersCam2[j].y, 2)));  // Euclidean
            }

            sqDist = homDist[j];
            if (sqDist < bestDist)
            {
                bestDist = sqDist;
                bestHomTmp = j;
            }

        }

        /* Compute the corresponding point in the other modality
        This includes the weighting of multiple homographies
        */
        vector<Point2f> tmpSourcePoint, tmpEstimatedPoint;
        Point2f estimatedPoint = { 0,0 };

        tmpSourcePoint.push_back(sourcePoints[i]);

        vector<int> quadrantIndicesTmp; vector<double> quadrantDistancesTmp;


        if (MAP_TYPE == 1) {
            bilinearInterpolator(sourcePoints[i], stereoCalibParam.homCentersCam1, homDist, homWeights,
                quadrantIndicesTmp, quadrantDistancesTmp);

            weightedHomographyMapper(tmpSourcePoint, tmpEstimatedPoint, stereoCalibParam.homCam1Cam2, homWeights);

        }
        else {
            bilinearInterpolator(sourcePoints[i], stereoCalibParam.homCentersCam2, homDist, homWeights,
                quadrantIndicesTmp, quadrantDistancesTmp);

            weightedHomographyMapper(tmpSourcePoint, tmpEstimatedPoint, stereoCalibParam.homCam2Cam1, homWeights);
        }

        estimatedPoint = tmpEstimatedPoint[0];

        if (MAP_TYPE == 1) {
            cam2Points.push_back(estimatedPoint);
        }
        else {
            cam1Points.push_back(estimatedPoint);
        }
    }
}


void Registrator::computeHomographyMapping(vector<Point2f>& vecUndistRgbCoord, vector<Point2f>& vecUndistTCoord, vector<Point2f> vecDCoord,
    vector<int> vecDepthInMm, vector<double>& minDist, vector<int> &bestHom, vector<vector<int> > &octantIndices,
    vector<vector<double> > &octantDistances, vector<Point3f> &worldCoordPointVector)
{
    // Use trilinearInterpolator

    double depthInMm, sqDist;
    Scalar sumOfWeights;
    vector<double> homDist, homWeights;
    vector<Point2f> tmpUndistRgbPoint, tmpUndistTPoint, tmpEstimatedPoint;
    Point2f estimatedPoint;
    Point tmpIdx;
    int MAP_TYPE = 0;
    int nbrCoordinates;
    int nbrClusters = int(stereoCalibParam.homDepthCentersRgb.size());
    int count = 0;
    int prevDepthInMm = stereoCalibParam.defaultDepth;

    // Are we mapping RGB or thermal points?
    if (vecUndistRgbCoord.size() > 0)
    {
        MAP_TYPE = 1;
        nbrCoordinates = int(vecUndistRgbCoord.size());
        vecUndistTCoord.clear();
    }
    else if (vecUndistTCoord.size() > 0)
    {
        MAP_TYPE = 2;
        nbrCoordinates = int(vecUndistTCoord.size());
        vecUndistRgbCoord.clear();
    }
    else
    {
        MAP_TYPE = 0;
        cerr << "Error in computeHomographyMapping - no coordinates to map!";
        return;
    }

    for (int i = 0; i < nbrCoordinates; i++)
    {
        if (vecDepthInMm.size() == 0) // Do we need to look up the depth?
        {
            if (vecDCoord.size() > 0)  // Are we providing any depth coordinates (as in the case of thermal coordinates)
            {
                if ((vecDCoord[i].y > 0) && (vecDCoord[i].x > 0) && (vecDCoord[i].y <= stereoCalibParam.HEIGHT)
                    && (vecDCoord[i].x <= stereoCalibParam.WIDTH))
                {
                    depthInMm = lookUpDepth(dImg, vecDCoord[i], true);

                    if (depthInMm > 7000)
                    { // Depth is undefined
                        depthInMm = prevDepthInMm;
                    }
                }
                else {
                    depthInMm = prevDepthInMm;
                }
            }
            else {
                depthInMm = prevDepthInMm;
            }
        }
        else
        {
            // If not, the depth has previously been determined
            depthInMm = float(vecDepthInMm[i]);
        }

        prevDepthInMm = int(depthInMm);

        // Find the euclidean distances from the point to the homographies
        Mat worldCoordPoint(1, 3, CV_32F, Scalar(0));
        Point3f tmpWorldCoord;


        // First, map the image coordinates to world coordinates
        if (MAP_TYPE == 1) {
            worldCoordPoint.at<float>(0, 0) = backProjectPoint(vecUndistRgbCoord[i].x, float(stereoCalibParam.rgbCamMat.at<double>(0, 0)),
                float(stereoCalibParam.rgbCamMat.at<double>(0, 2)), float(depthInMm));
            worldCoordPoint.at<float>(0, 1) = backProjectPoint(vecUndistRgbCoord[i].y, float(stereoCalibParam.rgbCamMat.at<double>(1, 1)),
                float(stereoCalibParam.rgbCamMat.at<double>(1, 2)), float(depthInMm));
            worldCoordPoint.at<float>(0, 2) = float(depthInMm);
        }
        else {
            worldCoordPoint.at<float>(0, 0) = backProjectPoint(vecUndistTCoord[i].x, float(stereoCalibParam.tCamMat.at<double>(0, 0)),
                float(stereoCalibParam.tCamMat.at<double>(0, 2)), 1500);
            worldCoordPoint.at<float>(0, 1) = backProjectPoint(vecUndistTCoord[i].y, float(stereoCalibParam.tCamMat.at<double>(1, 1)),
                float(stereoCalibParam.tCamMat.at<double>(1, 2)), 1500);
            worldCoordPoint.at<float>(0, 2) = 1500;

        }

        tmpWorldCoord.x = worldCoordPoint.at<float>(0, 0);
        tmpWorldCoord.y = worldCoordPoint.at<float>(0, 1);
        tmpWorldCoord.z = worldCoordPoint.at<float>(0, 2);
        worldCoordPointVector.push_back(tmpWorldCoord);

        int bestHomTmp = 0; // Reset parameters
        double bestDist = 1e6;
        homDist.clear(); homWeights.clear();

        for (int j = 0; j < nbrClusters; j++)
        {
            if (stereoCalibParam.homDepthCentersRgb[j].z >= 0) // Is the current cluster valid? (also applies for thermal clusters)
            {
                Mat depthCenter(1, 3, CV_32F, Scalar(0));
                // Compute Euclidean distance
                if (MAP_TYPE == 1)
                {
                    homDist.push_back(sqrt(pow(worldCoordPoint.at<float>(0, 0) - stereoCalibParam.homDepthCentersRgb[j].x, 2) +
                        pow(worldCoordPoint.at<float>(0, 1) - stereoCalibParam.homDepthCentersRgb[j].y, 2) +
                        pow(worldCoordPoint.at<float>(0, 2) - stereoCalibParam.homDepthCentersRgb[j].z, 2)));  // Euclidean					
                }
                else {
                    homDist.push_back(sqrt(pow(worldCoordPoint.at<float>(0, 1) - stereoCalibParam.homDepthCentersT[j].x, 2) +
                        pow(worldCoordPoint.at<float>(0, 1) - stereoCalibParam.homDepthCentersT[j].y, 2) +
                        pow(worldCoordPoint.at<float>(0, 2) - stereoCalibParam.homDepthCentersT[j].z, 2)));  // Euclidean
                }

                for (size_t k = 0; k < settings.discardedHomographies.size(); ++k) {
                    if (j == settings.discardedHomographies[k]) {
                        homDist[j] = 1e12;
                    }
                }

                sqDist = homDist[j];
                if (sqDist < bestDist)
                {
                    bestDist = sqDist;
                    bestHomTmp = j;
                }

            }
            else {
                homDist.push_back(1e12);
            }
        }

        /* Compute the corresponding point in the other modality
        This includes the weighting of multiple homographies
        */
        tmpUndistRgbPoint.clear(); tmpUndistTPoint.clear();
        estimatedPoint.x = 0; estimatedPoint.y = 0;
        if (MAP_TYPE == 1) {
            tmpUndistRgbPoint.push_back(vecUndistRgbCoord[i]);
        }
        else {
            tmpUndistTPoint.push_back(vecUndistTCoord[i]);
        }

        vector<int> octantIndicesTmp; vector<double> octantDistancesTmp;

        if (MAP_TYPE == 1) {
            trilinearInterpolator(tmpWorldCoord, stereoCalibParam.homDepthCentersRgb, homDist, homWeights,
                octantIndicesTmp, octantDistancesTmp);

            weightedHomographyMapper(tmpUndistRgbPoint, tmpEstimatedPoint, stereoCalibParam.planarHom, homWeights);

        }
        else {
            trilinearInterpolator(tmpWorldCoord, stereoCalibParam.homDepthCentersT, homDist, homWeights,
                octantIndicesTmp, octantDistancesTmp);

            weightedHomographyMapper(tmpUndistTPoint, tmpEstimatedPoint, stereoCalibParam.planarHomInv, homWeights);
        }

        octantIndices.push_back(octantIndicesTmp);
        octantDistances.push_back(octantDistancesTmp);
        estimatedPoint = tmpEstimatedPoint[0];

        if (MAP_TYPE == 1) {
            vecUndistTCoord.push_back(estimatedPoint);
        }
        else {
            vecUndistRgbCoord.push_back(estimatedPoint);
        }

        minDist.push_back(bestDist);
        bestHom.push_back(bestHomTmp);
    }

    if (count > 0) {
        std::cout << endl;
    }
}

void Registrator::trilinearInterpolator(Point3f inputPoint, vector<Point3f> &sourcePoints, vector<double> &precomputedDistance, vector<double> &weights,
    vector<int> &nearestSrcPointInd, vector<double> &nearestSrcPointDist)
{
    /*	TrilinearHomographyInterpolator finds the nearest point for each quadrant in 3D space and calculates weights
    based on trilinear interpolation for the input 3D point. The function returns a list of weights of the points
    used for the interpolation
    */

    /* Octant map:
    I:		+ + +
    II:		- + +
    III:	- - +
    IV:		+ - +
    V:		+ + -
    VI:		- + -
    VII:	- - -
    VIII:	+ - -
    */
    weights.clear();
    nearestSrcPointDist.clear();
    nearestSrcPointInd.clear();


    // Step 1: Label each sourcePoint according to the octant it belongs to
    vector<int> octantMap;

    for (size_t i = 0; i < sourcePoints.size(); ++i)
    {

        if ((sourcePoints[i].x - inputPoint.x) > 0)
        {	// x is positive; We are either in the first, fourth, fifth, or eighth octant

            if ((sourcePoints[i].y - inputPoint.y) > 0)
            { // y is positive; We are either in the first or fifth octant

                if ((sourcePoints[i].z - inputPoint.z) > 0)
                { // z is positive; We are in the first octant
                    octantMap.push_back(1);
                }
                else {
                    // z is negative; We are in the fifth octant				
                    octantMap.push_back(5);
                }
            }
            else {
                // y is negative; We are either in the fourth or eighth octant

                if ((sourcePoints[i].z - inputPoint.z) > 0)
                { // z is positive; We are in the fourth octant
                    octantMap.push_back(4);
                }
                else {
                    // z is negative; We are in the eighth octant				
                    octantMap.push_back(8);
                }
            }
        }
        else {
            // x is negative; We are either in the second, third, sixth, or seventh octant

            if ((sourcePoints[i].y - inputPoint.y) > 0)
            { // y is positive; We are either in the second or sixth octant

                if ((sourcePoints[i].z - inputPoint.z) > 0)
                { // z is positive; We are in the second octant
                    octantMap.push_back(2);
                }
                else {
                    // z is negative; We are in the sixth octant				
                    octantMap.push_back(6);
                }
            }
            else {
                // y is negative; We are either in the third or seventh octant

                if ((sourcePoints[i].z - inputPoint.z) > 0)
                { // z is positive; We are in the third octant
                    octantMap.push_back(3);
                }
                else {
                    // z is negative; We are in the seventh octant				
                    octantMap.push_back(7);
                }
            }
        }
    }

    // Step 2: Find the nearest point for every octant

    for (int i = 0; i < 8; ++i)
    {
        nearestSrcPointInd.push_back(-1);
        nearestSrcPointDist.push_back(1e6);
    }

    int currentOctant;
    double currentDist;

    for (size_t i = 0; i < sourcePoints.size(); ++i)
    {
        // Identify the current octant
        currentOctant = octantMap[i];

        // Identify the distance to the input point
        currentDist = precomputedDistance[i];

        if (nearestSrcPointDist[currentOctant - 1] > currentDist)
        {
            nearestSrcPointInd[currentOctant - 1] = i;
            nearestSrcPointDist[currentOctant - 1] = currentDist;
        }
    }

    // Step 3: Compute the relative weight of the eight surrounding points
    vector<double> unNormWeights;
    double tmpWeight;
    double sumOfDistances = 0;

    for (size_t i = 0; i < nearestSrcPointDist.size(); ++i)
    {
        if (nearestSrcPointDist[i] < 1e6)
            sumOfDistances += nearestSrcPointDist[i];
    }

    for (size_t i = 0; i < nearestSrcPointDist.size(); ++i)
    {
        if (nearestSrcPointDist[i] < 1e6)
        {
            tmpWeight = sumOfDistances / nearestSrcPointDist[i];
            unNormWeights.push_back(tmpWeight);
        }
        else {
            unNormWeights.push_back(0.);
        }
    }

    // Step 4: Normalize the weights
    double sumOfWeights = 0;
    vector<double> tmpWeightsVec;

    for (size_t i = 0; i < nearestSrcPointDist.size(); ++i)
    {
        sumOfWeights += unNormWeights[i];
    }

    for (size_t i = 0; i < nearestSrcPointDist.size(); ++i)
    {
        tmpWeight = unNormWeights[i] / sumOfWeights;
        tmpWeightsVec.push_back(tmpWeight);

        nearestSrcPointDist[i] = tmpWeight;
    }

    // Step 5: Recreate the weights to make a weight for every distance in precomputedDistance, even if the weight is zero
    for (size_t i = 0; i < precomputedDistance.size(); ++i)
    {
        weights.push_back(0.);
    }

    for (size_t i = 0; i < tmpWeightsVec.size(); ++i)
    {
        if (nearestSrcPointInd[i] >= 0)
        {
            weights[nearestSrcPointInd[i]] = tmpWeightsVec[i];
        }
    }

}

void Registrator::bilinearInterpolator(Point2f inputPoint, vector<Point2f> &sourcePoints, vector<double> &precomputedDistance, vector<double> &weights,
    vector<int> &nearestSrcPointInd, vector<double> &nearestSrcPointDist)
{
    /*	BilinearHomographyInterpolator finds the nearest point for each quadrant in 3D space and calculates weights
    based on bilinear interpolation for the input 2D point. The function returns a list of weights of the points
    used for the interpolation
    */

    /* Quadrant map:
    I:		+ +
    II:		- +
    III:	- -
    IV:		+ -
    */
    weights.clear();
    nearestSrcPointDist.clear();
    nearestSrcPointInd.clear();

    // If we have more than four sourcePoints, use bilinearInterpolation. Otherwise use a simple interpolation to avoid unlinearities

    if (sourcePoints.size() > 4)
    {
        // Step 1: Label each sourcePoint according to the quadrant it belongs to
        vector<int> quadrantMap;
        float distanceX, distanceY, distanceZ;

        for (auto i = 0; i < sourcePoints.size(); ++i)
        {
            if ((sourcePoints[i].x - inputPoint.x) > 0)
            {	// x is positive; We are either in the first or fourth oct

                if ((sourcePoints[i].y - inputPoint.y) > 0)
                { // y is positive; We are in the first quadrant

                    quadrantMap.push_back(1);
                }
                else {
                    // y is negative; We are in the fourth quadrant
                    quadrantMap.push_back(4);
                }
            }
            else {
                // x is negative; We are either in the second or third quadrant

                if ((sourcePoints[i].y - inputPoint.y) > 0)
                { // y is positive; We are in the second quadrant

                    quadrantMap.push_back(2);
                }
                else {
                    // y is negative; We are in the third quadrant
                    quadrantMap.push_back(3);
                }
            }
        }

        // Step 2: Find the nearest point for every quadrant

        for (int i = 0; i < 8; ++i)
        {
            nearestSrcPointInd.push_back(-1);
            nearestSrcPointDist.push_back(1e6);
        }

        int currentquadrant;
        double currentDist;

        for (auto i = 0; i < sourcePoints.size(); ++i)
        {
            // Identify the current quadrant
            currentquadrant = quadrantMap[i];

            // Identify the distance to the input point
            currentDist = precomputedDistance[i];

            if (nearestSrcPointDist[currentquadrant - 1] > currentDist)
            {
                nearestSrcPointInd[currentquadrant - 1] = i;
                nearestSrcPointDist[currentquadrant - 1] = currentDist;
            }
        }
    }
    else
    {
        // If we have four or less sourcePoints, compute the weights based on the distance from inputPoints and sourcePoints
        for (auto i = 0; i < sourcePoints.size(); ++i)
        {
            nearestSrcPointInd.push_back(i);
            nearestSrcPointDist.push_back(precomputedDistance[i]);
        }
    }

    // Step 3: Compute the relative weight of the four surrounding points
    vector<double> unNormWeights;
    double tmpWeight;
    double sumOfDistances = 0;

    for (auto i = 0; i < nearestSrcPointDist.size(); ++i)
    {
        if (nearestSrcPointDist[i] < 1e6)
            sumOfDistances += nearestSrcPointDist[i];
    }

    for (auto i = 0; i < nearestSrcPointDist.size(); ++i)
    {
        if (nearestSrcPointDist[i] < 1e6)
        {
            tmpWeight = sumOfDistances / nearestSrcPointDist[i];
            unNormWeights.push_back(tmpWeight);
        }
        else {
            unNormWeights.push_back(0.);
        }
    }

    // Step 4: Normalize the weights
    double sumOfWeights = 0;
    vector<double> tmpWeightsVec;

    for (auto i = 0; i < nearestSrcPointDist.size(); ++i)
    {
        sumOfWeights += unNormWeights[i];
    }

    for (auto i = 0; i < nearestSrcPointDist.size(); ++i)
    {
        tmpWeight = unNormWeights[i] / sumOfWeights;
        tmpWeightsVec.push_back(tmpWeight);

        nearestSrcPointDist[i] = tmpWeight;
    }

    // Step 5: Recreate the weights to make a weight for every distance in precomputedDistance, even if the weight is zero
    for (auto i = 0; i < precomputedDistance.size(); ++i)
    {
        weights.push_back(0.);
    }

    for (auto i = 0; i < tmpWeightsVec.size(); ++i)
    {
        if (nearestSrcPointInd[i] >= 0)
        {
            weights[nearestSrcPointInd[i]] = tmpWeightsVec[i];
        }
    }

}

void Registrator::weightedHomographyMapper(vector<Point2f> undistPoint, vector<Point2f> &estimatedPoint, vector<Mat> &homographies, vector<double> &homWeights)
{
    /* weightedHomographyMapper maps the undistPoint (undistorted point) based by a weighted sum of the provided homographies, which are weighted by homWeights
    */
    assert(homWeights.size() == homographies.size());
    vector<Point2f> tmpEstimatedPoint; Point2f tmpPoint;
    estimatedPoint.clear();
    estimatedPoint.push_back(tmpPoint);

    for (size_t i = 0; i < homWeights.size(); ++i)
    {
        if (homWeights[i] > 0)
        {
            perspectiveTransform(undistPoint, tmpEstimatedPoint, homographies[i]);

            estimatedPoint[0].x = estimatedPoint[0].x + tmpEstimatedPoint[0].x * float(homWeights[i]);
            estimatedPoint[0].y = estimatedPoint[0].y + tmpEstimatedPoint[0].y * float(homWeights[i]);
        }
    }

}

void Registrator::depthOutlierRemovalLookup(vector<Point2f> vecDCoord, vector<int> &vecDepthInMm)
{
    // depthOutlierRemovalLookup looks up depth for the provided coordinates of vecDCoord, discards garbage values
    // and makes constraints on the output depth in order to provide a smoothed depth lookup and removal of outlier values

    int depthInMm, depthNaNThreshold, validDepthMeasCount = 0, sumDepthInMm = 0;
    bool SCALE_TO_THEORETICAL = true;
    int avgDepthInMm;
    vecDepthInMm.clear();
    vector<int> rawDepthInMm;

    if (SCALE_TO_THEORETICAL) {
        depthNaNThreshold = int(stereoCalibParam.depthCoeffA * 6500 + stereoCalibParam.depthCoeffB);
    }
    else {
        depthNaNThreshold = 6500;
    }

    if (vecDCoord.size() > 0)
    {

        // Step 1: Look up the depth for the entire depth coordinate set
        for (size_t i = 0; i < vecDCoord.size(); ++i)
        {
            if (vecDCoord.size() > 0)  // Are we providing any depth coordinates (as in the case of thermal coordinates)
            {
                if ((vecDCoord[i].y > 0) && (vecDCoord[i].x > 0) && (vecDCoord[i].y <= stereoCalibParam.HEIGHT)
                    && (vecDCoord[i].x <= stereoCalibParam.WIDTH))
                {
                    depthInMm = int(lookUpDepth(dImg, vecDCoord[i], SCALE_TO_THEORETICAL));


                    if (depthInMm > depthNaNThreshold)
                    { // Depth is undefined
                        depthInMm = depthNaNThreshold;
                    }
                    else {
                        // We have a valid depth measurement
                        sumDepthInMm += depthInMm;
                        ++validDepthMeasCount;
                    }
                }
                else {
                    // Depth is undefined
                    depthInMm = depthNaNThreshold;
                }
            }
            else {
                cerr << "No depth coordinates provided";
            }

            rawDepthInMm.push_back(depthInMm);
        }

        // Step 2: Filter outliers. Identify the outliers by perfoming K-means clustering and filter them 
        // according to a threshold. As we assume, that the human contours are in the foreground, this must
        // correspond to the minimum depth measurement cluster. Other depth measurements should be within a range
        // this cluster, otherwise they may be classified as outliers


        Mat labels, centers;
        Mat depthMat = Mat::zeros(Size(1, rawDepthInMm.size()), CV_32F);
        int nbrClusters;

        if (validDepthMeasCount < vecDCoord.size()) {
            // If there exists undefined depth in our dataset, we will need an extra cluster
            // to contain this
            nbrClusters = settings.nbrClustersForDepthOutlierRemoval;
        }
        else {
            // Otherwise, we will go for two clusters
            nbrClusters = settings.nbrClustersForDepthOutlierRemoval - 1;
        }

        if (vecDCoord.size() > nbrClusters) {
            for (size_t i = 0; i < vecDCoord.size(); ++i) {
                depthMat.at<float>(i, 0) = float(rawDepthInMm[i]);
            }

            kmeans(depthMat, nbrClusters, labels, TermCriteria(CV_TERMCRIT_EPS, 10, 0.5), 100,
                KMEANS_RANDOM_CENTERS, centers);

            int minAvg = depthNaNThreshold;

            for (int i = 0; i < nbrClusters; ++i) {
                // Find the lowest mean, or centre
                if (centers.at<float>(i, 0) < minAvg) {
                    minAvg = centers.at<float>(i, 0);
                }
            }
            avgDepthInMm = minAvg;

            vector<int> clustersToBeDiscarded;

            for (int i = 0; i < nbrClusters; ++i) {
                // If depths are more than depthProximityThreshold mm away from the human contour, they should be regarded as noise
                if (centers.at<float>(i, 0) >(minAvg + settings.depthProximityThreshold)) {
                    clustersToBeDiscarded.push_back(i);
                }
            }

            for (size_t i = 0; i < vecDCoord.size(); ++i) {
                // Replace the noisy depth values by the average
                int currentCentre = labels.at<int>(i, 0);

                for (size_t j = 0; j < clustersToBeDiscarded.size(); ++j) {
                    if (currentCentre == clustersToBeDiscarded[j]) {
                        rawDepthInMm[i] = minAvg;
                    }
                }
            }

            vector<int> labelsVec, centersVec; // For debug
            for (size_t i = 0; i < vecDCoord.size(); ++i) {
                labelsVec.push_back(labels.at<int>(i, 0));
            }

            for (int i = 0; i < nbrClusters; ++i) {
                centersVec.push_back(int(centers.at<float>(i, 0)));
            }
        }
        else {
            // If we are operating on a very small dataset, calculate the mean manually
            if (validDepthMeasCount > 0) {
                avgDepthInMm = sumDepthInMm / validDepthMeasCount;
            }
            else {
                avgDepthInMm = stereoCalibParam.defaultDepth;
            }
        }

        // Step 3: Let the data through a non-causal moving-average filter
        int windowLength = 4;
        int filteredValue, validCounts;

        for (size_t i = 0; i < vecDCoord.size(); ++i)
        {
            if (rawDepthInMm[i] != depthNaNThreshold) { // Extract the current value
                filteredValue = rawDepthInMm[i];
                validCounts = 1;
            }
            else {
                filteredValue = 0;
                validCounts = 0;
            }

            // Go backwards
            int j = i - 1;
            int currentStep = 0;

            while ((currentStep < (windowLength / 2)) && (j >= 0))
            {
                if (rawDepthInMm[j] != depthNaNThreshold) {
                    // If we have a valid value, add it to the filtered value
                    filteredValue += rawDepthInMm[j];
                    validCounts++;
                    currentStep++;
                }
                --j;
            }

            // Go forwards
            j = i + 1;
            currentStep = 0;


            while ((currentStep < (windowLength / 2)) && (j < vecDCoord.size()))
            {
                if (rawDepthInMm[j] != depthNaNThreshold) {
                    // If we have a valid value, add it to the filtered value
                    filteredValue += rawDepthInMm[j];
                    validCounts++;
                    currentStep++;
                }
                ++j;
            }

            // Compute the smoothed value and insert it into the depth vector
            if (validCounts > 0) {
                vecDepthInMm.push_back(filteredValue / validCounts);
            }
            else {
                vecDepthInMm.push_back(avgDepthInMm);
            }
        }
    }
}

void Registrator::MyDistortPoints(std::vector<cv::Point2f> src, std::vector<cv::Point2f> & dst,
    const cv::Mat & cameraMatrix, const cv::Mat & distortionMatrix)
{
    // Normalize points before entering the distortion process
    Mat zeroDist;
    zeroDist = Mat::zeros(1, 5, CV_32F);
    std::vector<cv::Point2f> normalizedUndistPoints;
    undistortPoints(src, normalizedUndistPoints, cameraMatrix, zeroDist);

    src = normalizedUndistPoints;


    // Code thanks to http://stackoverflow.com/questions/10935882/opencv-camera-calibration-re-distort-points-with-camera-intrinsics-extrinsics
    dst.clear();
    float fx = float(cameraMatrix.at<double>(0, 0));
    float fy = float(cameraMatrix.at<double>(1, 1));
    float ux = float(cameraMatrix.at<double>(0, 2));
    float uy = float(cameraMatrix.at<double>(1, 2));

    float k1 = float(distortionMatrix.at<double>(0, 0));
    float k2 = float(distortionMatrix.at<double>(0, 1));
    float p1 = float(distortionMatrix.at<double>(0, 2));
    float p2 = float(distortionMatrix.at<double>(0, 3));
    float k3 = 0;

    if (distortionMatrix.size().width > 4) {
        k3 = float(distortionMatrix.at<double>(0, 4));
    }

    //BOOST_FOREACH(const cv::Point2d &p, src)
    for (unsigned int i = 0; i < src.size(); i++)
    {
        const cv::Point2f &p = src[i];
        float x = p.x;
        float y = p.y;
        float xCorrected, yCorrected;
        //Step 1 : correct distorsion
        {
            float r2 = x*x + y*y;
            //radial distorsion
            xCorrected = x * (1 + k1 * r2 + k2 * r2 * r2 + k3 * r2 * r2 * r2);
            yCorrected = y * (1 + k1 * r2 + k2 * r2 * r2 + k3 * r2 * r2 * r2);

            //tangential distorsion
            //The "Learning OpenCV" book is wrong here !!!
            //False equations from the "Learning OpenCv" book
            //xCorrected = xCorrected + (2. * p1 * y + p2 * (r2 + 2. * x * x)); 
            //yCorrected = yCorrected + (p1 * (r2 + 2. * y * y) + 2. * p2 * x);
            //Correct formulae found at : http://www.vision.caltech.edu/bouguetj/calib_doc/htmls/parameters.html
            xCorrected = xCorrected + (2 * p1 * x * y + p2 * (r2 + 2 * x * x));
            yCorrected = yCorrected + (p1 * (r2 + 2 * y * y) + 2 * p2 * x * y);
        }
        //Step 2 : ideal coordinates => actual coordinates
        {
            xCorrected = xCorrected * float(fx) + float(ux);
            yCorrected = yCorrected * float(fy) + float(uy);
        }
        dst.push_back(cv::Point2d(xCorrected, yCorrected));
    }

}

void Registrator::setImageWidth(int width)
{
    if (width > 0) {
        stereoCalibParam.WIDTH = width;
    }
}

void Registrator::setImageHeight(int height)
{
    if (height > 0) {
        stereoCalibParam.HEIGHT = height;
    }
}

bool Registrator::getIsInitialized()
{
    return settings.isInitialized;
}



void Registrator::loadMinCalibrationVars(string calFile)
{
    FileStorage fsStereo;

    try {
        fsStereo.open(calFile, FileStorage::READ);
    }
    catch (cv::Exception e) {
        QString errorMsg = "Exception happened opening file " + QString::fromStdString(calFile)
            + "\n\nException:\n" + QString::fromStdString(e.err) + "\n\nMessage: \n" + QString::fromStdString(e.msg);
        QMessageBox::critical(NULL, "Error opening calibration variables", errorMsg);
    }


    if (fsStereo.isOpened())
    { // If the file exists, we may read the stereo camera calibration parameters

        fsStereo["registrationMethod"] >> settings.registrationMethod;

        if (settings.registrationMethod == TrimodalMultipleHom)
        {
            // Depth to RGB registration maps
            fsStereo["rgbToDCalX"] >> stereoCalibParam.rgbToDCalX;
            fsStereo["rgbToDCalY"] >> stereoCalibParam.rgbToDCalY;
            fsStereo["dToRgbCalX"] >> stereoCalibParam.dToRgbCalX;
            fsStereo["dToRgbCalY"] >> stereoCalibParam.dToRgbCalY;

            // Homographies
            fsStereo["planarHom"] >> stereoCalibParam.planarHom;
            fsStereo["planarHomInv"] >> stereoCalibParam.planarHomInv;

            // 3D representations of the "centres" of the homographies
            fsStereo["homDepthCentersRgb"] >> stereoCalibParam.homDepthCentersRgb;
            fsStereo["homDepthCentersT"] >> stereoCalibParam.homDepthCentersT;

            // Misc parameters
            fsStereo["depthCoeffA"] >> stereoCalibParam.depthCoeffA;
            fsStereo["depthCoeffB"] >> stereoCalibParam.depthCoeffB;
            fsStereo["defaultDepth"] >> stereoCalibParam.defaultDepth;
            fsStereo["WIDTH"] >> stereoCalibParam.WIDTH;
            fsStereo["HEIGHT"] >> stereoCalibParam.HEIGHT;

            // Flags and settings
            fsStereo["UNDISTORT_IMAGES"] >> settings.UNDISTORT_IMAGES;
            fsStereo["depthProximityThreshold"] >> settings.depthProximityThreshold;
            fsStereo["nbrClustersForDepthOutlierRemoval"] >> settings.nbrClustersForDepthOutlierRemoval;

            if (settings.nbrClustersForDepthOutlierRemoval == 0)
            {
                settings.nbrClustersForDepthOutlierRemoval = 3;
            }

            fsStereo["discardedHomographies"] >> settings.discardedHomographies;
        }
        else if (settings.registrationMethod == BimodalMultipleHom)
        {
            fsStereo["homCam1Cam2"] >> stereoCalibParam.homCam1Cam2;
            fsStereo["homCam2Cam1"] >> stereoCalibParam.homCam2Cam1;
            fsStereo["nbrHom"] >> stereoCalibParam.nbrHom;
            fsStereo["homCentersCam1"] >> stereoCalibParam.homCentersCam1;
            fsStereo["homCentersCam2"] >> stereoCalibParam.homCentersCam2;
        }

        // Intrinsic camera parameters and distortion parameters
        fsStereo["cam1CamMat"] >> stereoCalibParam.rgbCamMat;
        fsStereo["cam2CamMat"] >> stereoCalibParam.tCamMat;

        if (stereoCalibParam.rgbCamMat.empty()) {
            stereoCalibParam.rgbCamMat = Mat::eye(Size(3, 3), CV_32FC1);
        }

        if (stereoCalibParam.tCamMat.empty()) {
            stereoCalibParam.tCamMat = Mat::eye(Size(3, 3), CV_32FC1);
        }

        fsStereo["cam1DistCoeff"] >> stereoCalibParam.rgbDistCoeff;
        fsStereo["cam2DistCoeff"] >> stereoCalibParam.tDistCoeff;

        bool distortionParametersLoaded = true;

        // Check the retrieved parameters. We should have at least 4 distortion parameters 
        // and at most 14 in order to use the OpenCV undistortion functions
        if (stereoCalibParam.rgbDistCoeff.size().area() < 4 || stereoCalibParam.rgbDistCoeff.size().area() > 14) {
            distortionParametersLoaded = false;

            // Choose sensible defaults
            stereoCalibParam.rgbDistCoeff = Mat::zeros(Size(1, 4), CV_32FC1);
        }

        if (stereoCalibParam.tDistCoeff.size().area() < 4 || stereoCalibParam.tDistCoeff.size().area() > 14) {
            distortionParametersLoaded = false;

            // Choose sensible defaults
            stereoCalibParam.tDistCoeff = Mat::zeros(Size(1, 4), CV_32FC1);
        }

        if (distortionParametersLoaded) {
            // If we have loaded the distortion parameters from the configuration file successfully, we 
            // should indicate, that we want to undistort the input coordinates, hence are 
            // the input images distorted
            settings.UNDISTORT_IMAGES = false;
        }
        else {
            // If not, we assume that we have loaded undistorted images or that we don't want to bother 
            // with rectifying the lens distortion
            settings.UNDISTORT_IMAGES = true;
        }

        settings.isInitialized = true;
    }
    fsStereo.release();
}

void Registrator::loadHomography(std::string calFile)
{
    FileStorage fs;

    try {
        fs.open(calFile, FileStorage::READ);
    }
    catch (cv::Exception e) {
        QString errorMsg = "Exception happened opening file " + QString::fromStdString(calFile)
            + "\n\nException:\n" + QString::fromStdString(e.err) + "\n\nMessage: \n" + QString::fromStdString(e.msg);
        QMessageBox::critical(NULL, "Error opening calibration variables", errorMsg);
    }

    settings.registrationMethod = BimodalSingleHom;

    if (fs.isOpened()) {
        fs["homRgbToT"] >> stereoCalibParam.homRgbToT;
        fs["homTToRgb"] >> stereoCalibParam.homTToRgb;

        // Check the acquired homographies
        if ((stereoCalibParam.homRgbToT.cols != 3) || (stereoCalibParam.homRgbToT.rows != 3) ||
            (stereoCalibParam.homTToRgb.cols != 3) || (stereoCalibParam.homTToRgb.rows != 3)) {
            QString errorMsg = QString("Could not load homographies from file. The homographies must be 3x3 matrices in .yml-format.\n\n"
                "Dimensions of loaded matrices:\nhomRgbToT: %1x%2\nhomTToRgb: %3x%3").arg(stereoCalibParam.homRgbToT.rows).arg(
                    stereoCalibParam.homRgbToT.cols).arg(stereoCalibParam.homTToRgb.rows).arg(stereoCalibParam.homTToRgb.cols);
            QMessageBox::critical(NULL, "Unable to acquire homographies from file", errorMsg);
        }
    }


}

void Registrator::drawRegisteredContours(cv::Mat &rgbContourImage, cv::Mat& depthContourImage, cv::Mat& thermalContourImage, cv::Mat depthImg, int mapType, 
    bool preserveColors)
{
    // This functions draws the contours in rgbContourImage in a registered fashion in depthContourImage and thermalContourImage.
    // Note that in the current version, this does not handle contours inside holes in contours - although one-level holes in
    // contours are handled just fine. This is due to CV_RETR_CCOMP. For complete support, CV_RETR_TREE should be used.

    vector< vector<Point> > contourPoints;
    vector<Vec4i> hierarchy;
    Mat workImg, sourceImg;

    switch (mapType) {
    case MAP_FROM_RGB:
    {
        workImg = rgbContourImage.clone();
        sourceImg = rgbContourImage;
        break;
    }
    case MAP_FROM_DEPTH:
    {
        workImg = depthContourImage.clone();
        sourceImg = depthContourImage;
        break;
    }
    case MAP_FROM_THERMAL:
    {
        workImg = thermalContourImage.clone();
        sourceImg = thermalContourImage;
        break;
    }
    default:
        workImg = rgbContourImage.clone();
        sourceImg = rgbContourImage;
    }

    Mat erodedContourImage = Mat::zeros(workImg.size(), workImg.type());
    Mat gradX, gradY, grad;
    Mat absGradX, absGradY;

    // Find the contours to draw:
    findContours(workImg, contourPoints, hierarchy, CV_RETR_CCOMP, CV_CHAIN_APPROX_SIMPLE);

    // Use the Sobel operator to find gradients of the image in order to shrink the contours
    Sobel(workImg, gradX, CV_32F, 1, 0, 3, 1, 0, BORDER_DEFAULT);
    Sobel(workImg, gradY, CV_32F, 0, 1, 3, 1, 0, BORDER_DEFAULT);


    // Draw them, one at a time:
    for (int i = 0; i < contourPoints.size(); i++) {

        // Shrink the contour to enhance the depth lookup. First, we need to compute the direction
        // of the intensity gradient from the Sobel derivatives
        vector<float> gradAngle;
        vector<Point> erodedContourPoints;
        int erosionSize = 5;
        int successCount = 0;
        Point2f prevPoint = contourPoints[i][0];

        for (size_t j = 0; j < contourPoints[i].size(); ++j) {
            float tmpAngle;
            tmpAngle = fastAtan2(gradX.at<float>(contourPoints[i][j]), gradY.at<float>(contourPoints[i][j]));
            gradAngle.push_back(tmpAngle);

            Point tmpErodedPoint;
            tmpErodedPoint.x = contourPoints[i][j].x - cos(tmpAngle)*erosionSize;
            tmpErodedPoint.y = contourPoints[i][j].y + sin(tmpAngle)*erosionSize;

            // Check if the eroded point is inside the mask
            if ((tmpErodedPoint.x < sourceImg.cols) && (tmpErodedPoint.y < sourceImg.rows) &&
                (tmpErodedPoint.x >= 0) && (tmpErodedPoint.y >= 0) &&
                (sourceImg.at<uchar>(tmpErodedPoint) != 0)) {
                erodedContourPoints.push_back(tmpErodedPoint);
                prevPoint = tmpErodedPoint;
                successCount++;
            }
            else {
                // Otherwise, use the previous point
                erodedContourPoints.push_back(prevPoint);
            }
        }

        vector<Point> rgbContour, dContour, tContour;
        getRegisteredContours(contourPoints[i], erodedContourPoints, rgbContour, dContour, tContour, depthImg, mapType);

        vector< vector<Point> > drawPoints;
        Scalar color;
        if (preserveColors) {
            color = Scalar::all(sourceImg.at<uchar>(contourPoints[i][0]));
        }
        else {
            color = Scalar::all(255); // Draw white as default.
        }
        if (hierarchy[i][3] != -1) { // Has parent. Thus, is a hole. Draw black.
            color = Scalar::all(0);
        }

        if (mapType == MAP_FROM_RGB || mapType == MAP_FROM_DEPTH) {
            drawPoints.push_back(tContour);
            drawContours(thermalContourImage, drawPoints, 0, color, CV_FILLED);
        }

        if (mapType == MAP_FROM_RGB || mapType == MAP_FROM_THERMAL) {
            drawPoints.clear();
            drawPoints.push_back(dContour);
            drawContours(depthContourImage, drawPoints, 0, color, CV_FILLED);
        }

        if (mapType == MAP_FROM_THERMAL || mapType == MAP_FROM_DEPTH) {
            drawPoints.clear();
            drawPoints.push_back(rgbContour);
            drawContours(rgbContourImage, drawPoints, 0, color, CV_FILLED);
        }
    }
}

void Registrator::saveRegisteredContours(Mat depthContourImage, Mat thermalContourImage, string depthSavePath, string thermalSavePath, string imgNbr)
{
    // This function saves the contours from drawRegisteredContours

    // Append the proper number and file extension for the save path
    depthSavePath.append(imgNbr);
    thermalSavePath.append(imgNbr);

    imwrite(depthSavePath, depthContourImage);
    imwrite(thermalSavePath, thermalContourImage);
}

void Registrator::buildContourDirectory(string rgbLoadPath, vector<string> &rgbIndices)
{
    // Load contour images in RGB folder using dirent.h
    DIR *dr;
    int count = 0, len;
    struct dirent *ent;
    const char * cRgbPath = rgbLoadPath.c_str();
    Mat tmpContourImg;

    dr = opendir(cRgbPath);
    if (dr != NULL) {
        while ((ent = readdir(dr)) != NULL) {
            // While we have not reached the end of the folder, proceed
            len = int(strlen(ent->d_name));
            if (len >= 4) {
                if (strcmp(".png", &(ent->d_name[len - 4])) == 0) {
                    rgbIndices.push_back(ent->d_name);
                }
            }
        }
    }

}

void Registrator::saveRegisteredImages(Mat depthImage, string depthSavePath, string imgNbr)
{
    // This function saves the contours from drawRegisteredContours

    // Append the proper number and file extension for the save path
    depthSavePath.append(imgNbr);
    imwrite(depthSavePath, depthImage);
}

void Registrator::buildImageDirectory(string rgbLoadPath, vector<string>& rgbIndices)
{
    string filetype = ".jpg";
    buildImageDirectory(rgbLoadPath, rgbIndices, filetype);
}

void Registrator::buildImageDirectory(string rgbLoadPath, vector<string>& rgbIndices, string filetype)
{
    DIR *dr;
    int count = 0, len;
    struct dirent *ent;
    const char * cRgbPath = rgbLoadPath.c_str();
    const char * fType = filetype.c_str();
    rgbIndices.clear();

    dr = opendir(cRgbPath);
    if (dr != NULL) {
        while ((ent = readdir(dr)) != NULL) {
            // While we have not reached the end of the folder, proceed
            len = int(strlen(ent->d_name));
            if (len >= 4) {
                if (strcmp(fType, &(ent->d_name[len - 4])) == 0) {
                    rgbIndices.push_back(ent->d_name);
                }
            }
        }
    }
}

void Registrator::transformRgbImageToDepth(Mat rgbSourceImg, Mat &dDestImg, vector<Point2f> rgbPoints, vector<Point2f> dPoints)
{

    for (size_t i = 0; i < dPoints.size(); ++i)
    {
        if (rgbPoints[i].x >= 1 && rgbPoints[i].x < rgbSourceImg.cols && rgbPoints[i].y >= 1 && rgbPoints[i].y < rgbSourceImg.rows)
        {
            dDestImg.at<uchar>(int(dPoints[i].y - 1), int(dPoints[i].x - 1)) = rgbSourceImg.at<uchar>(int(rgbPoints[i].y - 1), int(rgbPoints[i].x - 1));
        }

    }

}

void Registrator::transformDepthImageToRgb(Mat dSourceImg, Mat &rgbDestImg, vector<Point2f> dPoints, vector<Point2f> rgbPoints)
{
    for (size_t i = 0; i < rgbPoints.size(); ++i)
    {
        if (dPoints[i].x >= 1 && dPoints[i].x < dSourceImg.cols && dPoints[i].y >= 1 && dPoints[i].y < dSourceImg.rows)
        {
            rgbDestImg.at<ushort>(rgbPoints[i].y - 1, rgbPoints[i].x - 1) = dSourceImg.at<ushort>(dPoints[i].y - 1, dPoints[i].x - 1);
        }

    }

}


/// <summary> 
/// The rgbImagePath, depthImagePath must contain the full such as "E:/SyncRGB/"
/// </summary>

void Registrator::loadTransformRgb(string rgbImagePath, string depthImagePath)
{
    // This function loads RGB images, and one by one, registers them pixel-by-pixel to overlay them onto a depth image
    vector<string> rgbIndices;
    Mat rgbImage;

    buildImageDirectory(rgbImagePath, rgbIndices);
    vector<Point2f> dPoints, rgbPoints;

    for (size_t i = 0; i < rgbIndices.size(); ++i)
    {
        // Load the rgb image
        string tmpRgbPath = rgbImagePath;
        tmpRgbPath.append(rgbIndices[i]);
        rgbImage = imread(tmpRgbPath);

        if (i == 0)
        {
            for (size_t y = 1; y <= rgbImage.rows; ++y)
            {
                for (size_t x = 1; x <= rgbImage.cols; ++x)
                {
                    Point2f dPoint(x, y);
                    dPoints.push_back(dPoint);
                }
            }

            computeCorrespondingRgbPointFromDepth(dPoints, rgbPoints);
        }

        // Create a "registered" rgb image as if was captured by the depth camera
        Mat depthImage = Mat::zeros(rgbImage.rows, rgbImage.cols, CV_8UC3);
        transformRgbImageToDepth(rgbImage, depthImage, rgbPoints, dPoints);

        saveRegisteredImages(depthImage, depthImagePath, rgbIndices[i]);
        cout << i << " ";

    }

}

/// <summary> 
/// The rgbImagePath, depthImagePath must contain the full such as "E:/SyncRGB/"
/// </summary>

void Registrator::loadTransformDepth(string rgbImagePath, string depthImagePath)
{
    // This function loads RGB images, and one by one, registers them pixel-by-pixel to overlay them onto a depth image
    vector<string> depthIndices;
    Mat depthImage;

    buildImageDirectory(depthImagePath, depthIndices, ".png");
    vector<Point2f> dPoints, rgbPoints;

    for (size_t i = 0; i < depthIndices.size(); ++i)
    {
        // Load the depth image
        string tmpDepthPath = depthImagePath;
        tmpDepthPath.append(depthIndices[i]);
        depthImage = imread(tmpDepthPath, CV_LOAD_IMAGE_ANYDEPTH);

        if (i == 0)
        {
            for (int y = 1; y <= depthImage.rows; ++y)
            {
                for (int x = 1; x <= depthImage.cols; ++x)
                {
                    Point2f rgbPoint(x, y);
                    rgbPoints.push_back(rgbPoint);
                }
            }

            computeCorrespondingDepthPointFromRgb(rgbPoints, dPoints);
        }

        // Create a "registered" rgb image as if was captured by the depth camera
        Mat rgbImage = Mat::zeros(depthImage.rows, depthImage.cols, depthImage.type());
        transformDepthImageToRgb(depthImage, rgbImage, dPoints, rgbPoints);

        saveRegisteredImages(rgbImage, rgbImagePath, depthIndices[i]);
        std::cout << i << " ";

    }

}


/// <summary> 
/// The rgbContourPath, depthContourPath, thermalContourPath,  depthImagePath
/// must contain the full such as "E:/rgbMasks/"
/// </summary>


void Registrator::loadRegSaveContours(string rgbContourPath, string depthContourPath, string thermalContourPath,
    string depthImagePath)
{
    // This function loads a list of RGB contours, and, one by one, reads RGB contours, and 
    // registers them in the thermal and depth modalities

    vector<string> rgbIndices;
    Mat rgbContour, depthImage;

    this->buildContourDirectory(rgbContourPath, rgbIndices);

    for (size_t i = 0; i < rgbIndices.size(); ++i) {

        // Load the rgb contour and corresponding depth image
        string tmpRgbPath = rgbContourPath;
        tmpRgbPath.append(rgbIndices[i]);
        rgbContour = imread(tmpRgbPath, CV_LOAD_IMAGE_GRAYSCALE);

        string tmpDepthPath = depthImagePath;
        tmpDepthPath.append(rgbIndices[i]);
        depthImage = imread(tmpDepthPath, CV_LOAD_IMAGE_ANYDEPTH);

        // Create empty thermal and depth images
        Mat thermalContour = Mat::zeros(Size(stereoCalibParam.WIDTH, stereoCalibParam.HEIGHT), rgbContour.type());
        Mat	depthContour = Mat::zeros(Size(stereoCalibParam.WIDTH, stereoCalibParam.HEIGHT), rgbContour.type());

        // Register the thermal and depth contours
        drawRegisteredContours(rgbContour, depthContour, thermalContour, depthImage, true);

        //cvWaitKey(0);
        cout << rgbIndices[i] << ",";

        // Save the contours
        saveRegisteredContours(depthContour, thermalContour, depthContourPath, thermalContourPath, rgbIndices[i]);
    }
}

void Registrator::getRegisteredContours(vector<Point> contour, vector<Point> erodedContour, vector<Point>& rgbContour, 
    vector<Point>& dContour, vector<Point>& tContour, Mat depthImg, int mapType)
{
    vector<Point2f> rgbCoords, erodedRgbCoords, dCoords, erodedDCoords, tCoords, sourceCoords, erodedSourceCoords;
    depthImg.copyTo(this->dImg); // Required for the registration of RGB to thermal

    for (int j = 0; j < contour.size(); j++) {
        sourceCoords.push_back(Point(contour[j].x, contour[j].y));
    }

    for (int j = 0; j < erodedContour.size(); j++) {
        erodedSourceCoords.push_back(Point(erodedContour[j].x, erodedContour[j].y));
    }

    switch (settings.registrationMethod)
    {
    case TrimodalMultipleHom:
    {
        // Use advanced mapping supporting both RGB, depth, and thermal

        if (mapType == MAP_FROM_RGB || mapType == MAP_FROM_DEPTH) {

            if (mapType == MAP_FROM_RGB) {
                computeCorrespondingDepthPointFromRgb(erodedSourceCoords, erodedDCoords);
                computeCorrespondingDepthPointFromRgb(sourceCoords, dCoords);
                for (int j = dCoords.size() - 1; j >= 0; j--) {

                    if ((dCoords[j].x == 0) && (dCoords[j].y == 0)) {
                        // Depth coordinate is not valid. Remove corresponding rgbCoord and erodedDCoord
                        sourceCoords.erase(sourceCoords.begin() + j);
                        erodedDCoords.erase(erodedDCoords.begin() + j);
                    }
                    else {
                        // Depth coordinate is valid. Insert from beginning
                        dContour.insert(dContour.begin(), (Point(dCoords[j].x, dCoords[j].y)));
                    }
                }
            }
            else if (mapType == MAP_FROM_DEPTH) {
                computeCorrespondingRgbPointFromDepth(sourceCoords, rgbCoords);
                erodedDCoords = erodedSourceCoords;
            }

            vector<int> homInd, vecDepthInMm;
            vector<double> minDist;
            vector<vector<double>> octantDistances;
            vector<vector<int>> octantIndices;
            vector<Point3f> worldCoordPointVector;

            this->depthOutlierRemovalLookup(erodedDCoords, vecDepthInMm); // Use the coordinates of the eroded mask to look up the depth

            if (sourceCoords.size() > 0) {
                computeCorrespondingThermalPointFromRgb(sourceCoords, tCoords, dCoords, vecDepthInMm, minDist,
                    homInd, octantIndices, octantDistances, worldCoordPointVector);
            }
        }
        else if (mapType == MAP_FROM_THERMAL) {
            computeCorrespondingRgbPointFromThermal(sourceCoords, rgbCoords);
            computeCorrespondingDepthPointFromRgb(rgbCoords, dCoords);
        }


        break;
    }
    case BimodalMultipleHom:
    {
        // Use advanced mapping using both RGB and Thermal
        vector<Point2f> dCoords; // No  depth is provided

        if (mapType == MAP_FROM_RGB) {
            computeCorrespondingThermalPointFromRgb(sourceCoords, tCoords, dCoords);
        }
        else if (mapType == MAP_FROM_THERMAL) {
            computeCorrespondingRgbPointFromThermal(sourceCoords, rgbCoords);
        }

        break;
    }
    case BimodalSingleHom:
    {
        // Use a single, planar homography for the RGB -> thermal mapping

        if (mapType == MAP_FROM_RGB) {
            perspectiveTransform(sourceCoords, tCoords, stereoCalibParam.homRgbToT);
        }
        else if (mapType == MAP_FROM_THERMAL) {
            perspectiveTransform(sourceCoords, rgbCoords, stereoCalibParam.homTToRgb);
        }
        
        break;
    }
    }

    assert(stereoCalibParam.WIDTH != 0);
    assert(stereoCalibParam.HEIGHT != 0);

    for (auto i = 0; i < rgbCoords.size(); ++i) {
        Point2f adjustedCoord = rgbCoords[i];

        if (adjustedCoord.x < 0) {
            adjustedCoord.x = 0;
        }
        if (adjustedCoord.y < 0) {
            adjustedCoord.y = 0;
        }
        if (adjustedCoord.x >= stereoCalibParam.WIDTH) {
            adjustedCoord.x = stereoCalibParam.WIDTH - 1;
        }
        if (adjustedCoord.y >= stereoCalibParam.HEIGHT) {
            adjustedCoord.y = stereoCalibParam.HEIGHT - 1;
        }

        rgbContour.push_back(Point(adjustedCoord.x+0.5, adjustedCoord.y+0.5));
    }

    for (int j = 0; j < tCoords.size(); j++) {
        Point2f adjustedCoord = tCoords[j];

        if (adjustedCoord.x < 0) {
            adjustedCoord.x = 0;
        }
        if (adjustedCoord.y < 0) {
            adjustedCoord.y = 0;
        }
        if (adjustedCoord.x >= stereoCalibParam.WIDTH) {
            adjustedCoord.x = stereoCalibParam.WIDTH - 1;
        }
        if (adjustedCoord.y >= stereoCalibParam.HEIGHT) {
            adjustedCoord.y = stereoCalibParam.HEIGHT - 1;
        }

        tContour.push_back(Point(adjustedCoord.x+0.5, adjustedCoord.y+0.5));
    }

    if (dContour.empty()) {
        for (auto i = 0; i < dCoords.size(); ++i) {
            Point2f adjustedCoord = dCoords[i];

            if (adjustedCoord.x < 0) {
                adjustedCoord.x = 0;
            }
            if (adjustedCoord.y < 0) {
                adjustedCoord.y = 0;
            }
            if (adjustedCoord.x >= stereoCalibParam.WIDTH) {
                adjustedCoord.x = stereoCalibParam.WIDTH - 1;
            }
            if (adjustedCoord.y >= stereoCalibParam.HEIGHT) {
                adjustedCoord.y = stereoCalibParam.HEIGHT - 1;
            }

            dContour.push_back(Point(adjustedCoord.x+0.5, adjustedCoord.y+0.5));
        }
    }
}


// **** The remaining code is functions to show images and handle mouse callbacks in OpenCV
void Registrator::markCorrespondingPointsRgb(Point2f rgbCoord)
{
    vector<Point2f> vecTCoord, vecDCoord, epiTCoord, vecDistRgbCoord, vecRgbCoord, tempVecRgbCoord;
    Point2f tCoord, dCoord, distRgbCoord;
    Mat tmpDImg;
    vecRgbCoord.push_back(rgbCoord);
    vector<int> homInd;

    // ***** Compute the corresponding point in the depth image ******
    computeCorrespondingDepthPointFromRgb(vecRgbCoord, vecDCoord);
    dCoord = vecDCoord[0];

    // ***** Compute the corresponding point in the thermal image *****
    computeCorrespondingThermalPointFromRgb(vecRgbCoord, vecTCoord, vecDCoord, homInd);

    tCoord = vecTCoord[0];

    if (homInd.size() == 0)
    {
        homInd.push_back(0);
    }

    // Mark the points in the RGB, thermal, and depth image
    markCorrespondingPoints(rgbCoord, tCoord, dCoord, homInd[0], MAP_FROM_RGB);

}

void Registrator::markCorrespondingPointsThermal(Point2f tCoord)
{
    vector<Point2f> vecTCoord, vecRgbCoord, epiCoord, vecDCoord;
    Point2f rgbCoord, dCoord;
    vecTCoord.push_back(tCoord);
    Mat homEpiToRgb;


    // Find the corresponding point in RGB coordinates and process the information to markCorrespondingPoints()
    vector<double> minDist; vector<int> bestHom;  vector<vector<int>> octantIndices; vector<vector<double>> octantDistances;
    computeCorrespondingRgbPointFromThermal(vecTCoord, vecRgbCoord, minDist, bestHom, octantIndices, octantDistances);
    rgbCoord = vecRgbCoord[0];

    computeCorrespondingDepthPointFromRgb(vecRgbCoord, vecDCoord);
    dCoord = vecDCoord[0];

    if (bestHom.size() == 0)
    {
        bestHom.push_back(0);
    }

    markCorrespondingPoints(rgbCoord, tCoord, dCoord, bestHom[0], MAP_FROM_THERMAL);

}

void Registrator::markCorrespondingPointsDepth(Point2f dCoord)
{
    // Find the corresponding point in RGB coordinates and process the information to markCorrespondingPointsRgb()
    Point2f rgbCoord, tCoord;
    vector<Point2f> vecDCoord, vecRgbCoord, vecTCoord;
    vecDCoord.push_back(dCoord);

    computeCorrespondingRgbPointFromDepth(vecDCoord, vecRgbCoord);
    rgbCoord = vecRgbCoord[0];

    // ***** Compute the corresponding point in the thermal image *****
    vector<int> bestHom;

    computeCorrespondingThermalPointFromRgb(vecRgbCoord, vecTCoord, vecDCoord, bestHom);
    tCoord = vecTCoord[0];

    markCorrespondingPoints(rgbCoord, tCoord, dCoord, bestHom[0], MAP_FROM_DEPTH);
}

void Registrator::markCorrespondingPoints(Point2f rgbCoord, Point2f tCoord, Point2f dCoord, int homInd, int MAP_TYPE)
{
    Scalar rgbColor(0, 0, 0);
    Scalar tColor(0, 0, 0);
    Scalar dColor(65535, 65535, 65535);

    rgbColor[3] = 255;

    // Define the color of the circles based on the origin of the marking
    switch (MAP_TYPE)
    {
    case 0: // Map from RGB
        rgbColor[1] = 255;
        tColor[2] = 255;
        //dColor[2] = 255;
        break;
    case 1: // Map from thermal
        rgbColor[2] = 255;
        tColor[1] = 255;
        //dColor[2] = 255;
        break;
    case 2: // Map from depth
        rgbColor[2] = 255;
        tColor[2] = 255;
        //dColor[1] = 255;
        break;
    }

    // Mark the points in the RGB, thermal, and depth image
    circle(rgbImg, rgbCoord, 2, rgbColor, -1);
    imshow("RGB", rgbImg);

    circle(tImg, tCoord, 2, tColor, -1);
    imshow("Thermal", tImg);

    //Mat tmpDImg;
    //dImg.copyTo(tmpDImg);
    circle(dImg, dCoord, 2, dColor, -1);
    imshow("Depth", dImg);

    float depthInMm;
    // Find the depth value of the point and write it in the status window
    if ((dCoord.y > 0) && (dCoord.x > 0) && (dCoord.y <= stereoCalibParam.HEIGHT)
        && (dCoord.x <= stereoCalibParam.WIDTH))
    {
        depthInMm = lookUpDepth(dImg, dCoord, true);
    }
    else {
        depthInMm = 0;
    }


    // Write the coordinates in the status field
    char statusMsg[150];
    sprintf(statusMsg, "RGB: %3.0f,%3.0f :: Depth: %3.0f,%3.0f :: Thermal: %3.0f,%3.0f :: Depth of point: %4.1f mm :: Using hom #%d", rgbCoord.x, rgbCoord.y, dCoord.x, dCoord.y, tCoord.x, tCoord.y, depthInMm, homInd);

    displayStatusBar("RGB", statusMsg, 0);
}

void Registrator::updateFrames(int imgNbr, void* obj)
{
    Registrator* minReg = (Registrator*)obj; // Recast

    if (imgNbr == 0) {
        imgNbr = 1;
        setTrackbarPos("Frame", "RGB", imgNbr);
    }
    minReg->showModalityImages(minReg->rgbImgPath, minReg->tImgPath, minReg->dImgPath, imgNbr);
}

void Registrator::showModalityImages(string rgbPath, string tPath, string dPath, int imgNbr)
{
    string rgbImgStr, tImgStr, dImgStr;
    Mat distTImg, distRgbImg, rectTImg, rectRgbImg;
    Size imgSize(stereoCalibParam.WIDTH, stereoCalibParam.HEIGHT);

    // Insert the number of leading zeros
    stringstream nbrStr;
    nbrStr << setfill('0') << setw(5) << imgNbr;

    rgbImgStr = rgbPath + nbrStr.str() + ".jpg";
    tImgStr = tPath + nbrStr.str() + ".jpg";
    dImgStr = dPath + nbrStr.str() + ".png";

    distTImg = imread(tImgStr); distRgbImg = imread(rgbImgStr);  // Read thermal and rgb images

    try
    {
        if (settings.UNDISTORT_IMAGES)
        {
            undistort(distTImg, rectTImg, stereoCalibParam.tCamMat, stereoCalibParam.tDistCoeff);
            undistort(distRgbImg, rectRgbImg, stereoCalibParam.rgbCamMat, stereoCalibParam.rgbDistCoeff);
            tImg = rectTImg;
            rgbImg = rectRgbImg;
        }
        else {
            tImg = distTImg;
            rgbImg = distRgbImg;
        }
    }
    catch (exception& e) {
        cerr << "Unable to load images " << e.what() << endl;
    }

    try {
        imshow("Thermal", tImg); imshow("RGB", rgbImg); // Show thermal and rgb images
        dImg = imread(dImgStr, CV_LOAD_IMAGE_ANYDEPTH); imshow("Depth", dImg); // Read and show the depth image
    }
    catch (exception& e) {
        cerr << "Unable to show images " << e.what() << endl;
    }
}

void Registrator::createPointsLookupTable(vector<vector<Point2f> > vecOfVec, vector<int>& pointsToChessboardLookup)
{
    /* When the vector<vector<Point2f>> is concatenated to a single vector<Point2f>, the information of the placement of the
    point coordinates in the chessboard numbers is lost. This function provides a look-up-table that sets the range of the
    points for each of the chessboards
    */

    vector<Point2f> tmpVec;
    int vecSizeTmp = 0;
    pointsToChessboardLookup.clear();

    for (int i = 0; i < vecOfVec.size(); i++)
    {
        tmpVec = vecOfVec[i];
        vecSizeTmp = tmpVec.size() + vecSizeTmp;
        pointsToChessboardLookup.push_back(vecSizeTmp);
    }

}

void Registrator::vecOfVecToCatVec(vector<vector<Point2f> > vecOfVec, vector<Point2f>& catVec)
{
    for (auto i = 0; i < vecOfVec.size(); ++i)
    {
        catVec.insert(catVec.end(), vecOfVec[i].begin(), vecOfVec[i].end());
    }
}

void Registrator::vecOfVecToCatVec(vector<vector<double> > vecOfVec, vector<double>& catVec)
{
    for (auto i = 0; i < vecOfVec.size(); i++)
    {
        catVec.insert(catVec.end(), vecOfVec[i].begin(), vecOfVec[i].end());
    }
}

void Registrator::vecOfVecToCatVec(vector<vector<int> > vecOfVec, vector<int>& catVec)
{
    for (auto i = 0; i < vecOfVec.size(); i++)
    {
        catVec.insert(catVec.end(), vecOfVec[i].begin(), vecOfVec[i].end());
    }
}

void Registrator::vecOfVecToCatVec3D(vector<vector<Point3f> > vecOfVec, vector<Point3f>& catVec)
{
    for (auto i = 0; i < vecOfVec.size(); i++)
    {
        catVec.insert(catVec.end(), vecOfVec[i].begin(), vecOfVec[i].end());
    }
}

void Registrator::readChessboardSizeFromFile(string file, vector<Size>& calibrationPatternSize, vector<int>& activeImages)
{
    string line;
    char buff[20];
    int tempNum;
    ifstream fileH(file);
    int count = 0;
    calibrationPatternSize.clear(); activeImages.clear();

    if (fileH.is_open())
    {
        while (fileH.good())
        {
            // Clear temporary variables
            line.clear();
            count++;

            getline(fileH, line);
            stringstream subStr;
            subStr << line;

            while (subStr.good())
            {
                subStr.getline(buff, 5, ',');

                if (count == 1) {// If we are reading the first line, collect the indices of the active images


                    if (EOF != sscanf(buff, "%d", &tempNum)) {
                        activeImages.push_back(tempNum); // Place int in vector
                    }
                }
                else { // Otherwise, we need only to extract one number more and place it in the size vector

                    if (EOF != sscanf(buff, "%d", &tempNum))
                    {
                        int width = tempNum;
                        subStr.getline(buff, 5, ',');
                        sscanf(buff, "%d", &tempNum);
                        Size tempSize(width, tempNum);
                        calibrationPatternSize.push_back(tempSize);
                    }
                }
            }
        }
        fileH.close();
    }
    else cout << "Unable to open file";
}

void Registrator::computeRegistrationError(string valPointsPath, string valSavePath, string imgSavePath)
{
    // This function computes the RGB-to-thermal registration error by using the test and training data sets of the database. 
    // We maybe need to cover the registration errors in a manual fashion if the area covered by these samples is not enough to gain a broad representation of the registration

    // Create the error canvasses for different depths
    vector<Mat> sumOfTransferErrors, countOfTransferErrors, avgTransError;
    for (int i = 0; i < 3; ++i)
    {
        Mat sumCanvas = Mat::zeros(stereoCalibParam.HEIGHT, stereoCalibParam.WIDTH, CV_32F);
        Mat countCanvas = Mat::zeros(stereoCalibParam.HEIGHT, stereoCalibParam.WIDTH, CV_32F);
        Mat avgCanvas = Mat::zeros(stereoCalibParam.HEIGHT, stereoCalibParam.WIDTH, CV_32F);

        sumOfTransferErrors.push_back(sumCanvas);
        countOfTransferErrors.push_back(countCanvas);
        avgTransError.push_back(avgCanvas);
    }

    for (int folderIdx = 3; folderIdx <= 3; folderIdx++)
    {
        loadMinCalibrationVars("P:/Private/Trimodal registration/Dataset/Registration Variables/Scene " + to_string(folderIdx) + "/calibVars.yml");

        Mat tmpDepthImg, tmpTImg, tmpRectTImg, tmpRgbImg, tmpRectRgbImg;
        vector<int> pointsToChessboardLookup;
        vector<vector<Point2f>> rgbImagePoints, tImagePoints;

        // Step 1: Load the chessboard views. (This must be implemented somewhere in DualCalibration)
        vector<Point2f> rgbCatPoints, tCatPoints, rgbUndistPoints, tUndistPoints;
        for (int i = 1; i <= 3; ++i)
        {
            readCoordinatesFromFile2D(valPointsPath + to_string(folderIdx) + "/Cal/" + to_string(i) + "/CalibrationHelper/" + "lChessboard3D_imagec.txt", rgbImagePoints);
            readCoordinatesFromFile2D(valPointsPath + to_string(folderIdx) + "/Cal/" + to_string(i) + "/CalibrationHelper/" + "rChessboard3D_imagec.txt", tImagePoints);
        }

        createPointsLookupTable(rgbImagePoints, pointsToChessboardLookup);

        // The image points are now in vectors of vectors. Concatenate them into one, big vector
        vecOfVecToCatVec(rgbImagePoints, rgbCatPoints);
        vecOfVecToCatVec(tImagePoints, tCatPoints);

        // Undistort the points according to the distortion and camera matrices of the RGB and thermal cameras
        undistortPoints(rgbCatPoints, rgbUndistPoints, stereoCalibParam.rgbCamMat, stereoCalibParam.rgbDistCoeff,
            Mat::eye(3, 3, CV_32F), stereoCalibParam.rgbCamMat); // Undistort the points in the left image (rgb)
        undistortPoints(tCatPoints, tUndistPoints, stereoCalibParam.tCamMat, stereoCalibParam.tDistCoeff,
            Mat::eye(3, 3, CV_32F), stereoCalibParam.tCamMat); // Undistort the points in the right image (thermal)

        vector<Size> rCalibrationPatternSize;
        vector<int> activeImages;
        readChessboardSizeFromFile(valPointsPath + to_string(folderIdx) + "/Cal/" + "1/CalibrationHelper/" + "lChessboard3D_nbrCorners-3.txt", rCalibrationPatternSize, activeImages);

        vector<Mat> valDepthImg, valErrorMap;
        for (int k = 0; k < rgbImagePoints.size(); k++)
        {
            // Read all the calibration depth images into a buffer
            string dImgStr = valPointsPath + to_string(folderIdx) + "/Cal/" + "1/CalibrationHelper/D" + std::to_string(activeImages[k]) + ".png";
            tmpDepthImg = imread(dImgStr, CV_LOAD_IMAGE_ANYDEPTH);
            valDepthImg.push_back(tmpDepthImg);
        }

        // For each point, extract the corresponding depth point and the corresponding depth
        vector<Point2f> vecRgbCoord, vecDCoord; Point2f tmpPoint;
        vector<int> correspondingDepth;
        for (int i = 0; i < rgbUndistPoints.size(); i++)
        {
            if ((rgbUndistPoints[i].y <= stereoCalibParam.HEIGHT) && (rgbUndistPoints[i].y > 0)
                && (rgbUndistPoints[i].x <= stereoCalibParam.WIDTH) && (rgbUndistPoints[i].x > 0))
            {
                vecRgbCoord.push_back(rgbUndistPoints[i]);
                computeCorrespondingDepthPointFromRgb(vecRgbCoord, vecDCoord);
            }
            else {
                rgbUndistPoints[i].x = 0;
                rgbUndistPoints[i].y = 0;
                tUndistPoints[i].x = 0;
                tUndistPoints[i].y = 0;
                tmpPoint.x = -1; tmpPoint.y = -1;
                vecDCoord.push_back(tmpPoint);
            }

            // Find the chessboard number (and the validation image) that corresponds
            // to the extracted point
            int calImgNbr = 0;
            for (int j = 0; j < pointsToChessboardLookup.size(); j++)
            {
                if ((i) < pointsToChessboardLookup[j]) {
                    calImgNbr = j;
                    break;
                }
            }

            if ((vecDCoord[i].y <= stereoCalibParam.HEIGHT) && (vecDCoord[i].y > 0)
                && (vecDCoord[i].x <= stereoCalibParam.WIDTH) && (vecDCoord[i].x > 0))
            {
                int depthInMm;
                depthInMm = lookUpDepth(valDepthImg[calImgNbr], vecDCoord[i], true);

                if (depthInMm < 6000)
                {
                    correspondingDepth.push_back(depthInMm);
                }
                else {
                    correspondingDepth.push_back(stereoCalibParam.defaultDepth);
                }

            }
            else {
                correspondingDepth.push_back(stereoCalibParam.defaultDepth);
            }
            vecRgbCoord.clear();
        }


        // Step 2: Compute the registration error for each of the chessboard views. 
        vector<Point2f> mappedTCoord, mappedRgbCoord;
        vector<double> depthRectDisplacement, minDistRgb, minDistT, tmpMinDistT;
        vector<int> bestHomRgbToT, bestHomTToRgb, tmpBestHomTToRgb;
        vector<vector<double>> minDistTVecOfVec; vector<vector<int>> bestHomTToRgbVecOfVec;
        vector<vector<int>> octantIndicesRgbToT, octantIndicesTToRgb;
        vector<vector<double>> octantDistancesRgbToT, octantDistancesTToRgb;
        vector<Point3f> rgbWorldPoints, tWorldPoints;

        computeCorrespondingThermalPointFromRgb(rgbUndistPoints, mappedTCoord, vecDCoord, correspondingDepth, minDistRgb,
            bestHomRgbToT, octantIndicesRgbToT, octantDistancesRgbToT, rgbWorldPoints);


        vector<float> tmpErrorRgbToTX, tmpErrorRgbToTY, euclidTransErrorRgbToT;
        // Calculate errors btw. ground truth and mapping
        for (int i = 0; i < rgbUndistPoints.size(); i++)
        {
            int tmpErrorX = mappedTCoord[i].x - tUndistPoints[i].x;
            int tmpErrorY = mappedTCoord[i].y - tUndistPoints[i].y;

            if ((mappedTCoord[i].x >= 0) && (mappedTCoord[i].x <= stereoCalibParam.WIDTH) &&
                (mappedTCoord[i].y >= 0) && (mappedTCoord[i].y <= stereoCalibParam.HEIGHT))
            {
                tmpErrorRgbToTX.push_back(tmpErrorX);
                tmpErrorRgbToTY.push_back(tmpErrorY);
            }
            else {
                tmpErrorRgbToTX.push_back(0);
                tmpErrorRgbToTY.push_back(0);
            }
        }
        // Compute the Euclidean transfer error
        magnitude(tmpErrorRgbToTX, tmpErrorRgbToTY, euclidTransErrorRgbToT);

        // *******
        // Step 3: Decide on how to slice the information. Proposal: Provide an averaged view for 3 different slices of depth; i.e. of 1500-2000 mm, 2001-2500 mm, 2501 - 3000 mm. 
        //  ******

        // Assign the transfer errors to the corresponding depth canvas
        for (size_t i = 0; i < euclidTransErrorRgbToT.size(); ++i)
        {
            // Create the mask of points in close vicinity to the current point
            Mat mask = Mat::zeros(stereoCalibParam.HEIGHT, stereoCalibParam.WIDTH, CV_32F);
            // Draw a filled circle
            circle(mask, rgbUndistPoints[i], 30, Scalar(1), -1);

            // Find the right canvas
            //if (correspondingDepth[i] < 2000)
            //{
            // Depth is below 2000 mm. Put the error in the first canvas
            add(mask, countOfTransferErrors[0], countOfTransferErrors[0]);
            multiply(mask, euclidTransErrorRgbToT[i], mask);
            add(mask, sumOfTransferErrors[0], sumOfTransferErrors[0]);
            //}
            //else if (correspondingDepth[i] < 2500)
            //{
            //	// Depth is between 2500 - 3000 mm. Choose the second canvas
            //	add(mask, countOfTransferErrors[1], countOfTransferErrors[1]);
            //	multiply(mask, euclidTransErrorRgbToT[i], mask);
            //	add(mask, sumOfTransferErrors[1], sumOfTransferErrors[1]);
            //}
            //else {
            //	// Depth is above 2500 mm. Choose the third canvas
            //	add(mask, countOfTransferErrors[1], countOfTransferErrors[1]);
            //	multiply(mask, euclidTransErrorRgbToT[i], mask);
            //	add(mask, sumOfTransferErrors[1], sumOfTransferErrors[1]);
            //}
        }

    }

    vector<Mat> sumOfTransferErrorsU, countOfTransferErrorsU, avgTransErrorU;
    sumOfTransferErrorsU.resize(3); countOfTransferErrorsU.resize(3); avgTransErrorU.resize(3);

    for (int i = 0; i < 3; ++i)
    {
        sumOfTransferErrors[i].convertTo(sumOfTransferErrorsU[i], CV_16U);
        countOfTransferErrors[i].convertTo(countOfTransferErrorsU[i], CV_16U);
        imwrite("sumOfErrors" + std::to_string(i) + ".png", sumOfTransferErrorsU[i]);
        imwrite("countOfErrors" + std::to_string(i) + ".png", countOfTransferErrorsU[i]);

        divide(sumOfTransferErrors[i], countOfTransferErrors[i], avgTransError[i]);
        avgTransError[i].convertTo(avgTransErrorU[i], CV_16U);
        imwrite("avgTransError" + std::to_string(i) + ".png", avgTransErrorU[i]);
    }

    // Step 4: Choose a proper visualization: 
    // Proposal: A low visualization error is pure green, yellow is intermediate, red is large error. 
    // Proposal #2; Apply this in Matlab, use colormap()
    cout << "Finished";
}

void Registrator::rgbOnMouse(int event, int x, int y, int)
{
    if (event != CV_EVENT_LBUTTONDOWN)
        return;

    Point2i rgbCoord;
    rgbCoord.x = x;
    rgbCoord.y = y;

    markCorrespondingPointsRgb(rgbCoord);
}

void Registrator::thermalOnMouse(int event, int x, int y, int)
{
    if (event != CV_EVENT_LBUTTONDOWN)
        return;

    Point2i tCoord;
    tCoord.x = x;
    tCoord.y = y;

    markCorrespondingPointsThermal(tCoord);

}

void Registrator::depthOnMouse(int event, int x, int y, int)
{
    if (event != CV_EVENT_LBUTTONDOWN)
        return;

    Point2i dCoord;
    dCoord.x = x;
    dCoord.y = y;
    markCorrespondingPointsDepth(dCoord);

}

void Registrator::initWindows(int imgNbr)
{
    // Create windows
    namedWindow("RGB", CV_WINDOW_AUTOSIZE);
    namedWindow("Thermal", CV_WINDOW_AUTOSIZE);
    namedWindow("Depth", CV_WINDOW_AUTOSIZE);

    // Get number of images in rgb folder
    DIR *dr;
    int count = 0, len;
    struct dirent *pDirent;
    const char * cRgbPath = this->rgbImgPath.c_str();

    dr = opendir(cRgbPath);
    if (dr != NULL) {
        while ((pDirent = readdir(dr)) != NULL) {
            len = int(strlen(pDirent->d_name));
            if (len >= 4) {
                if (strcmp(".jpg", &(pDirent->d_name[len - 4])) == 0) {
                    count++;
                }
            }
        }
    }

    // Create trackbar
    cvCreateTrackbar2("Frame", "RGB", &imgNbr, count, Registrator::updateFrames, this);
    showModalityImages(rgbImgPath, tImgPath, dImgPath, imgNbr);
}

void wrappedRgbOnMouse(int event, int x, int y, int flags, void* ptr)
{
    Registrator* mcPtr = (Registrator*)ptr;
    if (mcPtr != NULL)
        mcPtr->rgbOnMouse(event, x, y, flags);
}

void wrappedThermalOnMouse(int event, int x, int y, int flags, void* ptr)
{
    Registrator* mcPtr = (Registrator*)ptr;
    if (mcPtr != NULL)
        mcPtr->thermalOnMouse(event, x, y, flags);
}

void wrappedDepthOnMouse(int event, int x, int y, int flags, void* ptr)
{
    Registrator* mcPtr = (Registrator*)ptr;
    if (mcPtr != NULL)
        mcPtr->depthOnMouse(event, x, y, flags);
}

string Registrator::getRgbImgPath()
{
    return this->rgbImgPath;
}

string Registrator::getTImgPath()
{
    return this->tImgPath;
}

string Registrator::getDImgPath()
{
    return this->dImgPath;
}

void Registrator::setRgbImgPath(string path)
{
    this->rgbImgPath = path;
}

void Registrator::setTImgPath(string path)
{
    this->tImgPath = path;
}

void Registrator::setDImgPath(string path)
{
    this->dImgPath = path;
}

void Registrator::toggleUndistortion(bool undistort)
{
    this->settings.UNDISTORT_IMAGES = undistort;
}

void Registrator::setDiscardedHomographies(vector<int> discardedHomographies)
{
    this->settings.discardedHomographies = discardedHomographies;
}

void Registrator::setUsePrevDepthPoint(bool value)
{
    this->settings.USE_PREV_DEPTH_POINT = value;
}